"""Represents a collection of data about a Source feature that is commonly used
to maintain and update SEA-PHAGES phage genomics data.
"""

from pdm_utils.functions import basic
from pdm_utils.classes import evaluation

class Source:

    def __init__(self):

        self.id = ""
        self.name = ""
        self.seqfeature = None # Biopython SeqFeature object.
        self.start = -1
        self.stop = -1
        self.organism = ""
        self.host = ""
        self.lab_host = ""

        # Data about the parent genome.
        self.genome_id = ""

        self._organism_name = "" # Parsed from organism.
        self._organism_host_genus = "" # Parsed from organism.
        self._host_host_genus = "" # Parsed from host.
        self._lab_host_host_genus = "" # Parsed from lab_host.
        self.evaluations = [] # List of warnings and errors about source feature.
        self.type = ""


    def parse_organism(self):
        """Retrieve the phage and host_genus names from the 'organism' field."""
        output = basic.parse_names_from_record_field(self.organism)
        self._organism_name = output[0]
        self._organism_host_genus = output[1]

    def parse_host(self):
        """Retrieve the host_genus name from the 'host' field."""
        self._host_host_genus = self.host.split(" ")[0]
        # Note: 'host' field is only expected to contain host information,
        # so a different strategy than parse_organism is used.

    def parse_lab_host(self):
        """Retrieve the host_genus name from the 'lab_host' field."""
        self._lab_host_host_genus = self.lab_host.split(" ")[0]
        # Note: 'host' field is only expected to contain host information,
        # so a different strategy than parse_organism is used.


    # Evaluations
    def set_eval(self, eval_id, definition, result, status):
        """Constructs and adds an Evaluation object to the evaluations list.

        :param eval_id: Unique identifier for the evaluation.
        :type eval_id: str
        :param definition: Description of the evaluation.
        :type definition: str
        :param result: Description of the outcome of the evaluation.
        :type result: str
        :param status: Outcome of the evaluation.
        :type status: str
        """
        evl = evaluation.Evaluation(eval_id, definition, result, status)
        self.evaluations.append(evl)


    def check_attribute(self, attribute, check_set, expect=False, eval_id=None,
                        success="correct", fail="error", eval_def=None):
        """Check that the attribute value is valid.

        :param attribute: Name of the Source feature object attribute to evaluate.
        :type attribute: str
        :param check_set: Set of reference ids.
        :type check_set: set
        :param expect:
            Indicates whether the attribute value is expected to be present
            in the check set.
        :type expect: bool
        :param eval_id: Unique identifier for the evaluation.
        :type eval_id: str
        :param success: Default status if the outcome is a success.
        :type success: str
        :param fail: Default status if the outcome is not a success.
        :type fail: str
        :param eval_def: Description of the evaluation.
        :type eval_def: str
        """
        try:
            test = True
            value1 = getattr(self, attribute)
        except:
            test = False
            value1 = None
        if test:
            value1_short = basic.truncate_value(str(value1), 30, "...")
            result = f"The {attribute} value '{value1_short}' is "
            value2 = basic.check_value_expected_in_set(
                        value1, check_set, expect)
            if value2:
                result = result + "valid."
                status = success
            else:
                result = result + "not valid."
                status = fail
        else:
            result = f"'{attribute}' is not a valid attribute to be evaluated."
            status = "untested"
        definition = f"Check the value of the '{attribute}' attribute."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)



import os
import shlex
from subprocess import Popen
import re

from pdm_utils.functions import basic


class AragornHandler:
    def __init__(self, identifier, sequence):
        self.id = identifier
        self.sequence = sequence

        # I/O attributes
        self.temp_dir = "/tmp/aragorn"
        self.input = os.path.join(self.temp_dir, self.id) + ".fasta"
        self.output = os.path.join(self.temp_dir, self.id) + ".out"
        self.out_str = ""

        self.trna_tally = 0
        self.tmrna_tally = 0

        self.trnas = list()
        self.tmrnas = list()

    def write_fasta(self):
        """
        Writes the search sequence to input file in FASTA format.
        :return:
        """
        if not os.path.exists(self.temp_dir):
            os.makedirs(self.temp_dir)
        with open(self.input, "w") as fh:
            fh.write(f">{self.id}\n{self.sequence}\n")

    def run_aragorn(self, c=False, d=True, m=False, t=True):
        """
        Set up Aragorn command, then run it. Default arguments will
        assume linear sequence to be scanned on both strands for tRNAs
        only (no tmRNAs).
        :param c: treat sequence as circular
        :type c: bool
        :param d: search both strands of DNA
        :type d: bool
        :param m: search for tmRNAs
        :type m: bool
        :param t: search for tRNAs
        :type t: bool
        :return:
        """
        command = f"aragorn -gcbact -br -wa "
        if c is False:
            command += "-l "
        if d is False:
            command += "-s "
        if m is True:
            command += "-m "
        if t is True:
            command += "-t "
        command += f"-o {self.output} {self.input}"

        Popen(shlex.split(command)).wait()

    def read_output(self):
        """
        Reads the Aragorn output file and joins the lines into a single
        string which it populates into `out_str`.
        :return:
        """
        with open(self.output, "r") as fh:
            self.out_str = "".join(fh.readlines())

    def parse_tmrnas(self):
        """
        Searches `out_str` for matches to a regular expression for
        Aragorn tmRNAs.
        :return:
        """
        # values:        orient,  start, stop,            peptide tag
        # indices:          0       1     2                    3
        re_str = "tmRNA\s+(\w+)?\[(-?\d+),(\d+)\]\s+\d+,\d+\s+([\w|*]*)"
        regex = re.compile(re_str, re.MULTILINE | re.DOTALL)

        tmrnas = regex.findall(self.out_str)

        # Iterate through tmRNAs and create dictionaries for each
        for tmrna in tmrnas:
            tmrna_data = dict()

            # Increment tally
            self.tmrna_tally += 1

            # If no complement character found, forward orientation
            if tmrna[0] == "":
                orientation = "forward"
            else:
                orientation = "reverse"

            # Coordinates need to be converted from 1-based closed to
            # 0-based half-open.
            start, stop = basic.reformat_coordinates(
                int(tmrna[1]), int(tmrna[2]), "1_closed", "0_half_open")

            tmrna_data["Orientation"] = orientation
            tmrna_data["Start"] = start
            tmrna_data["Stop"] = stop
            tmrna_data["PeptideTag"] = tmrna[3]

            self.tmrnas.append(tmrna_data)

    def parse_trnas(self):
        """
        Calls two helper methods to parse determinate and indeterminate
        tRNAs.
        :return:
        """
        self.parse_determinate_trnas()
        self.parse_indeterminate_trnas()

    def parse_determinate_trnas(self):
        """
        Searches `out_str` for matches to a regular expression for
        Aragorn tRNAs of determinate isotype.
        :return:
        """
        # values:       aa., orient, start, stop,          anticodon,
        # indices:       0      1      2     3                 4
        re_str = "tRNA-(\w+)\s+(c)?\[(-?\d+),(\d+)\]\s+\d+\s+\((\w+)\)\s+" \
                 "([.]*?\w+[.]*?)\s+([( )dAtv.]*)"
        # values:     sequence,      structure
        # indices:       5               6
        regex = re.compile(re_str, re.MULTILINE | re.DOTALL)

        trnas = regex.findall(self.out_str)

        # Iterate through tRNAs and create dictionaries for each
        for trna in trnas:
            trna_data = dict()

            # Increment tally
            self.trna_tally += 1

            amino_acid = trna[0]

            # If no complement character found, forward orientation
            if trna[1] == "":
                orientation = "forward"
            else:
                orientation = "reverse"

            # If start coordinate is negative, add 1 because Aragorn indexing
            # goes from -1 to 1 (skips 0)
            start, stop = int(trna[2]), int(trna[3])
            if start < 0:
                start += 1

            # Coordinates need to be converted from 1-based closed to
            # 0-based half open
            start, stop = basic.reformat_coordinates(
                start, stop, "1_closed", "0_half_open")

            anticodon = trna[4]
            sequence = trna[5]
            structure = trna[6]

            # Convert structure to true dot-bracket notation
            for char in [" ", "d", "A", "t", "v"]:
                structure = structure.replace(char, ".")

            while len(structure) < len(sequence):
                structure += "."

            trna_data["Orientation"] = orientation
            trna_data["Start"] = start
            trna_data["Stop"] = stop
            trna_data["AminoAcid"] = amino_acid
            trna_data["Anticodon"] = anticodon.lower()
            trna_data["Sequence"] = sequence
            trna_data["Structure"] = structure

            self.trnas.append(trna_data)

    def parse_indeterminate_trnas(self):
        """
        Searches `out_str` for matches to a regular expression for
        Aragorn tRNAs of indeterminate isotype.
        :return:
        """
        # values: possible amino acids., orient, start, stop,
        # indices:          0      1        2      3     4
        re_str = "tRNA-\?\((\w+)\|(\w+)\)\s+(c)?\[(-?\d+),(\d+)\]\s+\d+\s+" \
                 "\((\w+)\)\s+([.]*?\w+[.]*?)\s+([( )dAtv.]*)"
        # values: anticodon,      sequence,       structure
        # indices:    5              6                7
        regex = re.compile(re_str, re.MULTILINE | re.DOTALL)

        trnas = regex.findall(self.out_str)

        # Iterate through tRNAs and create dictionaries for each
        for trna in trnas:
            trna_data = dict()

            # Increment tally
            self.trna_tally += 1

            amino_acid = trna[0] + "|" + trna[1]

            # If no complement character found, forward orientation
            if trna[2] == "":
                orientation = "forward"
            else:
                orientation = "reverse"

            # If start coordinate is negative, add 1 because Aragorn indexing
            # goes from -1 to 1 (skips 0)
            start, stop = int(trna[3]), int(trna[4])
            if start < 0:
                start += 1

            # Coordinates need to be converted from 1-based closed to
            # 0-based half open
            start, stop = basic.reformat_coordinates(
                start, stop, "1_closed", "0_half_open")

            anticodon = trna[5]
            sequence = trna[6]
            structure = trna[7]

            # Convert structure to true dot-bracket notation
            for char in [" ", "d", "A", "t", "v"]:
                structure = structure.replace(char, ".")

            while len(structure) < len(sequence):
                structure += "."

            trna_data["Orientation"] = orientation
            trna_data["Start"] = start
            trna_data["Stop"] = stop
            trna_data["AminoAcid"] = amino_acid
            trna_data["Anticodon"] = anticodon.lower()
            trna_data["Sequence"] = sequence
            trna_data["Structure"] = structure

            self.trnas.append(trna_data)



"""Pipeline for exporting database information into files."""
import argparse
import shutil
import sys
import time
from pathlib import Path

from pdm_utils.classes.filter import Filter
from pdm_utils.functions import (configfile, fileio, flat_files, mysqldb,
                                 mysqldb_basic, pham_alignment,
                                 pipelines_basic, querying)


# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
DEFAULT_FOLDER_NAME = f"{time.strftime('%Y%m%d')}_export"
DEFAULT_TABLE = "phage"

TEMP_DIR = "/tmp/pdm_utils_export_temp"

PHAGE_QUERY = "SELECT * FROM phage"
GENE_QUERY = "SELECT * FROM gene"
TRNA_QUERY = "SELECT * FROM trna"
TMRNA_QUERY = "SELECT * FROM tmrna"

# Valid Biopython formats that crash the script due to specific values in
# some genomes that can probably be fixed relatively easily and implemented.
# BIOPYTHON_PIPELINES_FIXABLE = ["embl", "imgt", "seqxml"]
# Biopython formats that are not applicable.
# BIOPYTHON_PIPELINES_INVALID = ["fastq", "fastq-solexa", "fastq-illumina",
#                              "phd", "sff", "qual"]
# Biopython formats that are not writable
# BIOPYTHON_PIPELINES_NOT_WRITABLE = ["ig"]

# Valid file formats using Biopython
BIOPYTHON_PIPELINES = ["gb", "fasta", "clustal", "fasta-2line", "nexus",
                       "phylip", "pir", "stockholm", "tab"]
FILTERABLE_PIPELINES = BIOPYTHON_PIPELINES + ["csv", "tbl"]
PIPELINES = FILTERABLE_PIPELINES + ["sql"]
FLAT_FILE_TABLES = ["phage", "gene"]
FIVE_COLUMN_TABLES = ["phage"]

CDD_DATA_COLUMNS = ["gene_domain.QueryStart", "gene_domain.QueryEnd",
                    "domain.DomainID", "domain.Name", "domain.Description"]

# Once trna has data, these tables can be reintroduced.
TABLES = ["phage", "gene", "domain", "gene_domain", "pham",
          # "trna", "tmrna", "trna_structures",
          "version"]
SEQUENCE_COLUMNS = {"phage": ["Sequence"],
                    "gene": ["Translation"],
                    "domain": [],
                    "gene_domain": [],
                    "pham": [],
                    "trna": ["Sequence"],
                    "tmrna": [],
                    "trna_structures": [],
                    "version": []}


# MAIN FUNCTIONS
# -----------------------------------------------------------------------------

def main(unparsed_args_list):
    """Uses parsed args to run the entirety of the file export pipeline.

    :param unparsed_args_list: Input a list of command line args.
    :type unparsed_args_list: list[str]
    """
    # Returns after printing appropriate error message from parsing/connecting.
    args = parse_export(unparsed_args_list)

    config = configfile.build_complete_config(args.config_file)

    alchemist = pipelines_basic.build_alchemist(args.database, config=config)

    # Exporting as a SQL file is not constricted by schema version.
    if args.pipeline != "sql":
        mysqldb.check_schema_compatibility(alchemist.engine, "export")

    values = None
    if args.pipeline in FILTERABLE_PIPELINES:
        values = pipelines_basic.parse_value_input(args.input)
        if not values:
            values = None

    if args.pipeline not in PIPELINES:
        print("ABORTED EXPORT: Unknown pipeline option discrepency.\n"
              "Pipeline parsed from command line args is not supported")
        sys.exit(1)

    if args.pipeline != "I":
        execute_export(alchemist, args.pipeline, folder_path=args.folder_path,
                       folder_name=args.folder_name, table=args.table,
                       values=values, filters=args.filters, groups=args.groups,
                       sort=args.sort, include_columns=args.include_columns,
                       exclude_columns=args.exclude_columns,
                       sequence_columns=args.sequence_columns,
                       raw_bytes=args.raw_bytes,
                       concatenate=args.concatenate, db_name=args.db_name,
                       verbose=args.verbose, dump=args.dump, force=args.force,
                       threads=args.number_processes, phams_out=args.phams_out)
    else:
        pass


def parse_export(unparsed_args_list):
    """Parses export_db arguments and stores them with an argparse object.

    :param unparsed_args_list: Input a list of command line args.
    :type unparsed_args_list: list[str]
    :returns: ArgParse module parsed args.
    """
    EXPORT_SELECT_HELP = """
        Select a export pipeline option to export database data.
            Select csv to export data from a table into a .csv file.
            Select sql to dump the current database into a .sql file.
            Select a formatted file option to export individual entries.
        """
    DATABASE_HELP = "Name of the MySQL database to export from."

    CONFIG_FILE_HELP = """
        Export option that enables use of a config file for sourcing
        credentials
            Follow selection argument with the path to the config file
            specifying MySQL and NCBI credentials.
        """
    DUMP_HELP = """
        Export option that dumps exported files directly to the desired
        working directory.
        """
    FORCE_HELP = """
        Export option that aggresively creates and overwrites directories.
        """
    VERBOSE_HELP = """
        Export option that enables progress print statements.
        """
    FOLDER_PATH_HELP = """
        Export option to change the path
        of the directory where the exported files are stored.
            Follow selection argument with the path to the
            desired export directory.
        """
    FOLDER_NAME_HELP = """
        Export option to change the name
        of the directory where the exported files are stored.
            Follow selection argument with the desired name.
        """

    IMPORT_FILE_HELP = """
        Selection input option that imports values from a csv file.
            Follow selection argument with path to the
            csv file containing the names of each genome in the first column.
        """
    SINGLE_GENOMES_HELP = """
        Selection input option that imports values from cmd line input.
            Follow selection argument with space separated
            names of genomes in the database.
        """
    TABLE_HELP = """
        Selection option that changes the table in the database selected from.
            Follow selection argument with a valid table from the database.
        """
    WHERE_HELP = """
        Data filtering option that filters data by the inputted expressions.
            Follow selection argument with formatted filter expression:
                {Table}.{Column}={Value}
        """
    GROUP_BY_HELP = """
        Data selection option that groups data by the inputted columns.
            Follow selection argument with formatted column expressions:
                {Table}.{Column}={Value}
        """
    ORDER_BY_HELP = """
        Data selection option that sorts data by the inputted columns.
            Follow selection argument with formatted column expressions:
                {Table}.{Column}={Value}
        """

    DB_NAME_HELP = """
        MySQL export option to allow renaming of the exported database.
            Follow selection argument with the name of the desired database.
        """

    CONCATENATE_HELP = """
        SeqRecord export option to toggle concatenation of files.
            Toggle to enable concatenation of files.
        """

    SEQUENCE_COLUMNS_HELP = """
        Csv export option to toggle removal of sequence-based data.
            Toggle to include sequence based data.
        """
    INCLUDE_COLUMNS_HELP = """
        Csv export option to add additional columns of a database
        to an exported csv file.
            Follow selection argument with formatted column expressions:
                {Table}.{Column}={Value}
        """
    EXCLUDE_COLUMNS_HELP = """
        Csv export option to exclude select columns of a database
        from an exported csv file.
            Follow selection argument with formatted column expressions:
                {Table}.{Column}={Value}
        """
    RAW_BYTES_HELP = """
        Csv export option to conserve blob and encoded data from the database
        when exporting to a csv file.
            Toggle to leave conserve format of bytes-type data.
        """

    parser = argparse.ArgumentParser()
    parser.add_argument("database", type=str, help=DATABASE_HELP)

    subparsers = parser.add_subparsers(dest="pipeline", required=True,
                                       help=EXPORT_SELECT_HELP)

    filterable_parsers = []
    biopython_parsers = []
    subparser_list = []
    for pipeline in BIOPYTHON_PIPELINES:
        bio_parser = (subparsers.add_parser(pipeline))
        biopython_parsers.append(bio_parser)
        filterable_parsers.append(bio_parser)
        subparser_list.append(bio_parser)

    tbl_parser = subparsers.add_parser("tbl")
    csv_parser = subparsers.add_parser("csv")
    sql_parser = subparsers.add_parser("sql")

    subparser_list.append(tbl_parser)
    subparser_list.append(csv_parser)
    subparser_list.append(sql_parser)

    filterable_parsers.append(tbl_parser)
    filterable_parsers.append(csv_parser)

    for subparser in subparser_list:
        subparser.add_argument("-c", "--config_file",
                               type=pipelines_basic.convert_file_path,
                               help=CONFIG_FILE_HELP)
        subparser.add_argument("-m", "--folder_name", type=str,
                               help=FOLDER_NAME_HELP)
        subparser.add_argument("-o", "--folder_path", type=Path,
                               help=FOLDER_PATH_HELP)
        subparser.add_argument("-v", "--verbose", action="store_true",
                               help=VERBOSE_HELP)
        subparser.add_argument("-d", "--dump", action="store_true",
                               help=DUMP_HELP)
        subparser.add_argument("-f", "--force", action="store_true",
                               help=FORCE_HELP)
        subparser.add_argument("-np", "--number_processes", type=int)

    for subparser in filterable_parsers:
        table_choices = dict.fromkeys(BIOPYTHON_PIPELINES, FLAT_FILE_TABLES)
        table_choices["csv"] = TABLES
        table_choices["tbl"] = FIVE_COLUMN_TABLES

        subparser.add_argument("-if", "--import_file",
                               type=pipelines_basic.convert_file_path,
                               help=IMPORT_FILE_HELP, dest="input",
                               default=[])
        subparser.add_argument("-in", "--import_names", nargs="*",
                               help=SINGLE_GENOMES_HELP, dest="input")
        subparser.add_argument("-w", "--where", nargs="?",
                               help=WHERE_HELP, dest="filters")
        subparser.add_argument("-g", "--group_by", nargs="*",
                               help=GROUP_BY_HELP, dest="groups")
        subparser.add_argument("-s", "--order_by", nargs="*",
                               help=ORDER_BY_HELP)

    for subparser in biopython_parsers:
        subparser.add_argument("-cc", "--concatenate", help=CONCATENATE_HELP,
                               action="store_true")
        subparser.add_argument("-t", "--table", help=TABLE_HELP,
                               choices=FLAT_FILE_TABLES)

    tbl_parser.add_argument("-t", "--table", help=TABLE_HELP,
                            choices=FIVE_COLUMN_TABLES)

    csv_parser.add_argument("-t", "--table", help=TABLE_HELP,
                            choices=TABLES)
    csv_parser.add_argument("-sc", "--sequence_columns",
                            help=SEQUENCE_COLUMNS_HELP, action="store_true")
    csv_parser.add_argument("-ic", "--include_columns", nargs="*",
                            help=INCLUDE_COLUMNS_HELP)
    csv_parser.add_argument("-ec", "--exclude_columns", nargs="*",
                            help=EXCLUDE_COLUMNS_HELP)
    csv_parser.add_argument("-rb", "--raw_bytes", help=RAW_BYTES_HELP,
                            action="store_true")

    sql_parser.add_argument("-n", "--db_name", type=str, help=DB_NAME_HELP)
    sql_parser.add_argument("-pho", "--phams_out", action="store_true")

    for subparser in subparser_list:
        subparser.set_defaults(
                        folder_name=DEFAULT_FOLDER_NAME, folder_path=None,
                        config_file=None, verbose=False, input=[],
                        table=DEFAULT_TABLE, filters="", groups=[], sort=[],
                        include_columns=[], exclude_columns=[],
                        sequence_columns=False, concatenate=False,
                        raw_bytes=False, db_name=None, phams_out=False,
                        number_processes=1)

    parsed_args = parser.parse_args(unparsed_args_list[2:])

    return parsed_args


def execute_export(alchemist, pipeline, folder_path=None,
                   folder_name=DEFAULT_FOLDER_NAME, values=None, verbose=False,
                   dump=False, force=False, table=DEFAULT_TABLE, filters="",
                   groups=[], sort=[], include_columns=[], exclude_columns=[],
                   sequence_columns=False, raw_bytes=False, concatenate=False,
                   db_name=None, phams_out=False, threads=1):
    """Executes the entirety of the file export pipeline.

    :param alchemist: A connected and fully built AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param pipeline: File type that dictates data processing.
    :type pipeline: str
    :param folder_path: Path to a valid dir for new dir creation.
    :type folder_path: Path
    :param folder_name: A name for the export folder.
    :type folder_name: str
    :param force: A boolean to toggle aggresive building of directories.
    :type force: bool
    :param values: List of values to filter database results.
    :type values: list[str]
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    :param dump: A boolean value to toggle dump in current working dir.
    :type dump: bool
    :param table: MySQL table name.
    :type table: str
    :param filters: A list of lists with filter values, grouped by ORs.
    :type filters: str
    :param groups: A list of supported MySQL column names to group by.
    :type groups: list[str]
    :param sort: A list of supported MySQL column names to sort by.
    :type sort: list[str]
    :param include_columns: A csv export column selection parameter.
    :type include_columns: list[str]
    :param exclude_columns: A csv export column selection parameter.
    :type exclude_columns: list[str]
    :param sequence_columns: A boolean to toggle inclusion of sequence data.
    :type sequence_columns: bool
    :param concatenate: A boolean to toggle concaternation for SeqRecords.
    :type concaternate: bool
    :param threads: Number of processes/threads to spawn during the pipeline
    :type threads: int
    """
    if verbose:
        print("Retrieving database version...")
    db_version = mysqldb_basic.get_first_row_data(alchemist.engine, "version")

    if pipeline == "csv":
        if verbose:
            print("Processing columns for csv export...")
        csv_columns = filter_csv_columns(alchemist, table,
                                         include_columns=include_columns,
                                         exclude_columns=exclude_columns,
                                         sequence_columns=sequence_columns)

    if pipeline in FILTERABLE_PIPELINES:
        db_filter = pipelines_basic.build_filter(alchemist, table, filters,
                                                 values=values,
                                                 verbose=verbose)
        if sort:
            pipelines_basic.add_sort_columns(db_filter, sort, verbose=verbose)

    if verbose:
        print("Creating export folder...")
    export_path = pipelines_basic.create_working_path(folder_path, folder_name,
                                                      dump=dump, force=force)

    data_cache = {}
    if pipeline == "sql":
        execute_sql_export(alchemist, export_path, folder_path, db_version,
                           db_name=db_name, dump=dump, force=force,
                           phams_out=phams_out, threads=threads,
                           verbose=verbose)
    elif pipeline in FILTERABLE_PIPELINES:
        conditionals_map = pipelines_basic.build_groups_map(
                                                db_filter, export_path,
                                                groups=groups,
                                                verbose=verbose, force=force)

        if verbose:
            print("Prepared query and path structure, beginning export...")

        values = db_filter.values
        for mapped_path in conditionals_map.keys():
            db_filter.reset()
            db_filter.values = values

            conditionals = conditionals_map[mapped_path]
            db_filter.values = db_filter.build_values(where=conditionals)

            if db_filter.hits() == 0:
                print(f"No database entries received from {table} "
                      f"for '{mapped_path}'.")
                continue

            if sort:
                sort_columns = get_sort_columns(alchemist, sort)
                db_filter.sort(sort_columns)

            export_name = None
            if dump:
                if mapped_path == export_path:
                    export_name = folder_name

            pipelines_basic.create_working_dir(mapped_path, dump=dump,
                                               force=force)

            if pipeline in BIOPYTHON_PIPELINES + ["tbl"]:
                execute_ffx_export(alchemist, mapped_path, export_path,
                                   db_filter.values, pipeline, db_version,
                                   table, concatenate=concatenate,
                                   data_cache=data_cache,
                                   export_name=export_name, threads=threads,
                                   verbose=verbose, dump=dump)
            elif pipeline == "csv":
                execute_csv_export(db_filter, mapped_path, export_path,
                                   csv_columns, table, raw_bytes=raw_bytes,
                                   data_cache=data_cache,
                                   verbose=verbose, dump=dump)
    else:
        print("Unrecognized export pipeline, aborting export")
        sys.exit(1)


def execute_csv_export(db_filter, export_path, folder_path, columns, csv_name,
                       data_cache=None, sort=[], raw_bytes=False,
                       verbose=False, dump=False):
    """Executes csv export of a MySQL database table with select columns.

    :param db_filter: A connected and fully built Filter object.
    :type db_filter: Filter
    :param export_path: Path to a dir for file creation.
    :type export_path: Path
    :param folder_path: Path to a top-level dir.
    :type folder_path: Path
    :param table: MySQL table name.
    :type table: str
    :param conditionals: MySQL WHERE clause-related SQLAlchemy objects.
    :type conditionals: list[BinaryExpression]
    :param sort: A list of SQLAlchemy Columns to sort by.
    :type sort: list[Column]
    :param values: List of values to fitler database results.
    :type values: list[str]
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    :param dump: A boolean value to toggle dump in current working dir.
    :type dump: bool
    """
    if data_cache is None:
        data_cache = {}

    if verbose:
        relative_path = str(export_path.relative_to(folder_path))
        print(f"Preparing {csv_name} export for '{relative_path}'...")

    headers = [db_filter._key.name]
    for column in columns:
        if column.name != db_filter._key.name:
            headers.append(column.name)

    results = db_filter.select(columns)

    if not raw_bytes:
        decode_results(results, columns, verbose=verbose)

    if len(results) == 0:
        print(f"No database entries received for {csv_name}.")
        if not dump:
            export_path.rmdir()
    else:
        if dump:
            if export_path != folder_path:
                export_path.rmdir()
                export_path = export_path.parent
        if verbose:
            print(f"...Writing csv {csv_name}.csv in '{export_path.name}'...")

        file_path = export_path.joinpath(f"{csv_name}.csv")
        fileio.export_data_dict(results, file_path, headers,
                                include_headers=True)


def execute_ffx_export(alchemist, export_path, folder_path, values,
                       file_format, db_version, table, concatenate=False,
                       data_cache=None, verbose=False, dump=False,
                       threads=1, export_name=None):
    """Executes SeqRecord export of the compilation of data from a MySQL entry.

    :param alchemist: A connected and fully build AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param export_path: Path to a dir for file creation.
    :type export_path: Path
    :param folder_path: Path to a top-level dir.
    :type folder_path: Path
    :param file_format: Biopython supported file type.
    :type file_format: str
    :param db_version: Dictionary containing database version information.
    :type db_version: dict
    :param table: MySQL table name.
    :type table: str
    :param values: List of values to fitler database results.
    :type values: list[str]
    :param conditionals: MySQL WHERE clause-related SQLAlchemy objects.
    :type conditionals: list[BinaryExpression]
    :param sort: A list of SQLAlchemy Columns to sort by.
    :type sort: list[Column]
    :param concatenate: A boolean to toggle concatenation of SeqRecords.
    :type concaternate: bool
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    """
    if data_cache is None:
        data_cache = {}

    if export_name is None:
        export_name = export_path.name

    if verbose:
        print(f"Retrieving {export_name} data...")

    if table == "phage":
        seqrecords = get_genome_seqrecords(alchemist, values,
                                           data_cache=data_cache,
                                           verbose=verbose)
    elif table == "gene":
        seqrecords = get_cds_seqrecords(alchemist, values,
                                        data_cache=data_cache, verbose=verbose,
                                        file_format=file_format)
    else:
        print(f"Unknown error occured, table '{table}' is not recognized "
              "for SeqRecord export pipelines.")
        sys.exit(1)

    if file_format == "tbl":
        fileio.write_feature_table(seqrecords, export_path, verbose=verbose)
    else:
        if verbose:
            print("Appending database version...")
        for record in seqrecords:
            append_database_version(record, db_version)
        fileio.write_seqrecords(seqrecords, file_format, export_path,
                                export_name=export_name, verbose=verbose,
                                concatenate=concatenate, threads=threads)


def execute_sql_export(alchemist, export_path, folder_path, db_version,
                       db_name=None, dump=False, force=False, phams_out=False,
                       threads=1, verbose=False):
    pipelines_basic.create_working_dir(export_path, dump=dump, force=force)

    if phams_out:
        temp_dir = Path(TEMP_DIR)
        if temp_dir.is_dir():
            shutil.rmtree(temp_dir)
        temp_dir.mkdir()

        phams_out_fasta_dir = temp_dir.joinpath("fastas")
        pipelines_basic.create_working_dir(phams_out_fasta_dir,
                                           dump=dump, force=force)
        phams_out_aln_dir = temp_dir.joinpath("alns")
        pipelines_basic.create_working_dir(phams_out_aln_dir,
                                           dump=dump, force=force)

        phams_dict = pham_alignment.get_all_pham_gene_translations(alchemist)

        if verbose:
            print("...Writing and aligning pham fasta files...")
        pham_alignment.write_phams(phams_out_fasta_dir, phams_out_aln_dir,
                                   phams_dict, cores=threads, verbose=verbose)

        pham_fastas_zip = export_path.joinpath("fastas.zip")
        pham_alns_zip = export_path.joinpath("alns.zip")

        shutil.make_archive(pham_fastas_zip.with_suffix(""), "zip",
                            temp_dir, phams_out_fasta_dir.name)
        shutil.make_archive(pham_alns_zip.with_suffix(""), "zip",
                            temp_dir, phams_out_aln_dir.name)

    if verbose:
        print("Writing SQL database file...")

    fileio.write_database(alchemist, db_version["Version"], export_path,
                          db_name=db_name)


# EXPORT-SPECIFIC HELPER FUNCTIONS
# -----------------------------------------------------------------------------

# TODO Document and Unittest
def get_genome_seqrecords(alchemist, values, data_cache=None, verbose=False):
    if data_cache is None:
        data_cache = {}

    seqrecords = []
    for genome_id in values:
        genome = data_cache.get(genome_id)
        if genome is None:
            genome = get_single_genome(alchemist, genome_id, get_features=True,
                                       data_cache=data_cache)

        seqrecord = flat_files.genome_to_seqrecord(genome)
        flat_files.sort_seqrecord_features(seqrecord)
        seqrecords.append(seqrecord)

    return seqrecords


# TODO Document and Unittest
def get_cds_seqrecords(alchemist, values, data_cache=None, nucleotide=False,
                       verbose=False, file_format=None):
    if data_cache is None:
        data_cache = {}

    cds_list = parse_feature_data(alchemist, values=values)

    db_filter = Filter(alchemist)
    db_filter.key = 'gene.GeneID'

    if verbose:
        print("...Converting SQL data...")

    seqrecords = []
    for cds in cds_list:
        parent_genome = data_cache.get(cds.genome_id)

        if parent_genome is None:
            parent_genome = get_single_genome(alchemist, cds.genome_id,
                                              data_cache=data_cache)

        cds.genome_length = parent_genome.length
        cds.set_seqfeature()

        db_filter.values = [cds.id]
        gene_domains = db_filter.select(CDD_DATA_COLUMNS)

        record = flat_files.cds_to_seqrecord(cds, parent_genome,
                                             gene_domains=gene_domains,
                                             desc_type=file_format)
        seqrecords.append(record)

    return seqrecords


def get_sort_columns(alchemist, sort_inputs):
    """Function that converts input for sorting to SQLAlchemy Columns.

    :param alchemist: A connected and fully build AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param sort_inputs: A list of supported MySQL column names.
    :type sort_inputs: list[str]
    :returns: A list of SQLAlchemy Column objects.
    :rtype: list[Column]
    """
    sort_columns = []
    for sort_input in sort_inputs:
        try:
            sort_column = querying.get_column(alchemist.metadata, sort_input)
        except ValueError:
            print("Error occured while selecting sort columns.")
            print(f"Column inputted, '{sort_input}', is invalid.")
            sys.exit(1)
        finally:
            sort_columns.append(sort_column)

    return sort_columns


# CSV-EXPORT HELPER FUNCTIONS
# -----------------------------------------------------------------------------
def filter_csv_columns(alchemist, table, include_columns=[],
                       exclude_columns=[], sequence_columns=False):
    """Function that filters and constructs a list of Columns to select.

    :param alchemist: A connected and fully built AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param table: MySQL table name.
    :type table: str
    :param include_columns: A list of supported MySQL column names.
    :type include_columns: list[str]
    :param exclude_columns: A list of supported MySQL column names.
    :type exclude_columns: list[str]
    :param sequence_columns: A boolean to toggle inclusion of sequence data.
    :type sequence_columns: bool
    :returns: A list of SQLAlchemy Column objects.
    :rtype: list[Column]
    """
    table_obj = alchemist.metadata.tables[table]
    starting_columns = list(table_obj.columns)
    primary_key = list(table_obj.primary_key.columns)[0]

    include_column_objs = starting_columns
    for column in include_columns:
        try:
            column_obj = querying.get_column(alchemist.metadata, column)
        except ValueError:
            print("Error occured while selecting csv columns.")
            print(f"Column inputted, '{column}', is invalid.")
            sys.exit(1)
        finally:
            if column_obj not in include_column_objs:
                include_column_objs.append(column_obj)

    sequence_column_objs = []
    if not sequence_columns:
        for sequence_column in SEQUENCE_COLUMNS[table]:
            sequence_column_obj = dict(table_obj.c)[sequence_column]
            sequence_column_objs.append(sequence_column_obj)

    exclude_column_objs = sequence_column_objs
    for column in exclude_columns:
        try:
            column_obj = querying.get_column(alchemist.metadata, column)
        except ValueError:
            print("Error occured while selecting csv columns.")
            print(f"Column inputted, '{column}', is invalid.")
            sys.exit(1)
        finally:
            exclude_column_objs.append(column_obj)
            if column_obj.compare(primary_key):
                print(f"Primary key to {table} cannot be excluded")
                sys.exit(1)

            if column_obj not in exclude_column_objs:
                exclude_column_objs.append(column_obj)

    columns = []
    for column_obj in include_column_objs:
        if column_obj not in exclude_column_objs:
            columns.append(column_obj)

    return columns


def decode_results(results, columns, verbose=False):
    """Function that decodes encoded results from SQLAlchemy generated data.

    :param results: List of data dictionaries from a SQLAlchemy results proxy.
    :type results: list[dict]
    :param columns: SQLAlchemy Column objects.
    :type columns: list[Column]
    """
    for column in columns:
        if column.type.python_type == bytes:
            if verbose:
                print(f"Decoding retrieved {column} data...")
            for result in results:
                if not result[column.name] is None:
                    result[column.name] = result[column.name].decode("utf-8")


# Functions to be evaluated for another module:
# -----------------------------------------------------------------------------

# Copy of mysqldb.parse_feature_data() with value subquerying.
# Value subquerying needed for +300000 GeneID entries.
# Unsure about function redundancy.
def parse_feature_data(alchemist, values=[], limit=8000):
    """Returns Cds objects containing data parsed from a MySQL database.

    :param alchemist: A connected and fully built AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param values: List of GeneIDs upon which the query can be conditioned.
    :type values: list[str]
    """
    gene_table = querying.get_table(alchemist.metadata, "gene")
    primary_key = list(gene_table.primary_key.columns)[0]
    cds_data_columns = list(gene_table.c)

    cds_data_query = querying.build_select(alchemist.graph, cds_data_columns)

    cds_data = querying.execute(alchemist.engine, cds_data_query,
                                in_column=primary_key, values=values,
                                limit=limit)

    cds_list = []
    for data_dict in cds_data:
        cds_ftr = mysqldb.parse_gene_table_data(data_dict)
        cds_list.append(cds_ftr)

    return cds_list


def append_database_version(genome_seqrecord, version_data):
    """Function that appends the database version to the SeqRecord comments.

    :param genome_seqrecord: Filled SeqRecord object.
    :type genome_seqfeature: SeqRecord
    :param version_data: Dictionary containing database version information.
    :type version_data: dict
    """
    version_keys = version_data.keys()
    version = "NULL"
    schema_version = "NULL"
    if "Version" in version_keys or "SchemaVersion" not in version_keys:
        version = version_data["Version"]
    if "SchemaVersion" in version_keys:
        schema_version = version_data["SchemaVersion"]

    genome_seqrecord.annotations["comment"] =\
        genome_seqrecord.annotations["comment"] + (
            "Database Version: {}; Schema Version: {}".format(
                                                version, schema_version),)


# TODO Document
# TODO Evaluate for redundancy
def get_single_genome(alchemist, phageid, get_features=False, data_cache=None):
    gene_query = None
    trna_query = None
    tmrna_query = None
    if get_features:
        gene_query = GENE_QUERY
        trna_query = TRNA_QUERY
        tmrna_query = TMRNA_QUERY

    genome = mysqldb.parse_genome_data(
                            alchemist.engine, phage_id_list=[phageid],
                            phage_query=PHAGE_QUERY, gene_query=gene_query,
                            trna_query=trna_query, tmrna_query=tmrna_query)[0]

    if data_cache is not None:
        data_cache[phageid] = genome

    return genome



from collections import OrderedDict

from networkx import Graph
from networkx import shortest_path
from sqlalchemy import and_
from sqlalchemy import Column
from sqlalchemy import join
from sqlalchemy import MetaData
from sqlalchemy import select
from sqlalchemy import Table
from sqlalchemy.orm.decl_api import DeclarativeMeta
from sqlalchemy.sql import func
from sqlalchemy.sql import functions
from sqlalchemy.sql.elements import BinaryExpression
from sqlalchemy.sql.elements import BooleanClauseList
from sqlalchemy.sql.elements import Grouping
from sqlalchemy.sql.elements import Label
from sqlalchemy.sql.elements import UnaryExpression

from pdm_utils.functions import basic
from pdm_utils.functions import parsing

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------

COLUMN_TYPES = [Column, Table, functions.count, BinaryExpression,
                UnaryExpression, Label, DeclarativeMeta, Grouping]


# SQLALCHEMY OBJECT RETRIEVAL
# Functions that functionalize retrieval of SqlAlchemy objects.
# -----------------------------------------------------------------------------
def get_table(metadata, table):
    """Get a SQLAlchemy Table object, with a case-insensitive input.

    :param metadata: Reflected SQLAlchemy MetaData object.
    :type metadata: MetaData
    :param table: Case-insensitive table name.
    :type table: str
    :returns:
        Corresponding SQLAlchemy Table object.
    :rtype: Table
    """
    # Ensures table name is accurate before indexing metadata.
    table = parsing.translate_table(metadata, table)
    table_obj = metadata.tables[table]
    return table_obj


def get_column(metadata, column):
    """Get a SQLAlchemy Column object, with a case-insensitive input.
    Input must be formatted {Table_name}.{Column_name}.

    :param metadata: Reflected SQLAlchemy MetaData object.
    :type metadata: MetaData
    :param table: Case-insensitive column name.
    :type table: str
    :returns: Corresponding SQLAlchemy Column object.
    :rtype: Column
    """
    parsed_column = parsing.parse_column(column)

    table_obj = get_table(metadata, parsed_column[0])  # Retrieves SQL Table.
    # Ensures column name is accurate before indexing Table.
    column = parsing.translate_column(metadata, column)

    columns_dict = dict(table_obj.columns)  # Converts to indexable object.
    column_obj = columns_dict[column]

    return column_obj


# MySQL DATABASE MAPPING AND TRAVERSAL
# Functions to traverse MySQL database structures and map valid JOIN clauses.
# -----------------------------------------------------------------------------
def build_graph(metadata):
    """Get a NetworkX Graph object populated from a SQLAlchemy MetaData object.

    :param metadata: Reflected SQLAlchemy MetaData object.
    :type metadata: MetaData
    :returns: Populated and structured NetworkX Graph object.
    :rtype: Column
    """
    if not isinstance(metadata, MetaData):
        raise TypeError("Graph requires MetaData object, "
                        f"object passed was of type {type(metadata)}.")

    graph = Graph(metadata=metadata)  # Stores metadata in Graph object.
    for table in metadata.tables.keys():  # Creates nodes from all Tables.
        table_object = metadata.tables[table]
        graph.add_node(table_object.name, table=table_object)

    # Stores Table objects in Nodes
    for target_table in metadata.tables.keys():
        target_table_obj = metadata.tables[target_table]

        # Uses the Table object foreign keys to create Graph edges.
        for foreign_key in target_table_obj.foreign_keys:
            referent_column = foreign_key.column  # Retreives source Column
            referent_table_obj = referent_column.table
            referent_table = referent_table_obj.name

            # Stores the foreign key as a table name key pair [Target][Source]
            graph.add_edge(target_table, referent_table, key=foreign_key)

    return graph


def build_fromclause(db_graph, columns):
    """Get a joined table from pathing instructions for joining MySQL Tables.
    :param db_graph: SQLAlchemy structured NetworkX Graph object.
    :type db_graph: Graph
    :param columns: SQLAlchemy Column object(s).
    :type columns: Column
    :type columns: list
    :returns: SQLAlchemy Table object containing left outer-joined tables.
    :rtype: Table
    """
    table_list = get_table_list(columns)
    table_pathing = get_table_pathing(db_graph, table_list)
    joined_table = join_pathed_tables(db_graph, table_pathing)

    return joined_table


def build_onclause(db_graph, source_table, adjacent_table):
    """Creates a SQLAlchemy BinaryExpression object for a MySQL ON clause
       expression

    :param db_graph: SQLAlchemy structured NetworkX Graph object.
    :type db_graph: Graph
    :param source_table: Case-insensitive MySQL table name.
    :type source_table: str
    :param adjacent_table: Case-insensitive MySQL table name.
    :type source_table: str
    :returns: MySQL foreign key related SQLAlchemy BinaryExpression object.
    :rtype: BinaryExpression
    """
    # Retrieves accurate table names
    source_table = parsing.translate_table(db_graph.graph["metadata"],
                                           source_table)
    adjacent_table = parsing.translate_table(db_graph.graph["metadata"],
                                             adjacent_table)

    if source_table == adjacent_table:
        source_table = db_graph.nodes[source_table]["table"]
        return source_table

    edge = db_graph[source_table][adjacent_table]  # Indexes for graph edge.
    foreign_key = edge["key"]  # Retreives the stored foreign key object

    referent_column = foreign_key.column  # Gets the original primary key
    referenced_column = foreign_key.parent  # Gets the foreign key

    onclause = referent_column == referenced_column

    return onclause


def get_table_list(columns):
    """Get a nonrepeating list SQLAlchemy Table objects from Column objects.

    :param columns: SQLAlchemy Column object(s).
    :type columns: Column
    :type columns: list
    :returns: List of corresponding SQLAlchemy Table objects.
    :rtype: list
    """
    table_list = []

    if not isinstance(columns, list):
        columns = [columns]

    for column in columns:
        table_list.append(extract_column(column).table)

    # OrderedDict is used to remove repeats, trading space for runtime
    table_list = list(OrderedDict.fromkeys(table_list))

    return table_list


def get_table_pathing(db_graph, table_list, center_table=None):
    """Get pathing instructions for joining MySQL Table objects.

    :param db_graph: SQLAlchemy structured NetworkX Graph object.
    :type db_graph: Graph
    :param table_list: List of SQLAlchemy Table objects.
    :type table_list: list[Table]
    :param center_table: SQLAlchemy Table object to begin traversals from.
    :type center_table: Table
    :returns: 2-D list containing the center table and pathing instructions.
    :rtype: list
    """
    table_list = table_list.copy()

    # Removes the center_table from the list, if specified
    if center_table is not None:
        table_list.remove(center_table)
    # Creates a random center_table
    else:
        center_table = table_list[0]
        table_list.remove(center_table)

    table_paths = []
    for table in table_list:
        # NetworkX traversal algorithm finds path
        path = shortest_path(db_graph, center_table.name, table.name)

        if not path:
            raise ValueError("Operation cannot be performed. "
                             f"Table {table.name} is not connected by any "
                             f"foreign keys to table {center_table}.")

        table_paths.append(path)

    table_pathing = [center_table, table_paths]
    return table_pathing


def join_pathed_tables(db_graph, table_pathing):
    """Get a joined table from pathing instructions for joining MySQL Tables.

    :param db_graph: SQLAlchemy structured NetworkX Graph object.
    :type db_graph: Graph
    :param table_pathing: 2-D list containing a Table and pathing lists.
    :type table_pathing: list
    :returns: SQLAlchemy Table object containing left outer-joined tables.
    :rtype: Table
    """
    center_table = table_pathing[0]  # The center_table from table pathing
    joined_tables = center_table  # Creates a Table 'base' to add to

    joined_table_set = set()
    joined_table_set.add(center_table.name)  # Creates running table history

    for path in table_pathing[1]:
        for index in range(len(path)):
            table = path[index]
            previous_table = path[index-1]

            # If the table has not already been added:
            if table not in joined_table_set:
                joined_table_set.add(table)
                table_object = db_graph.nodes[table]["table"]
                # Build a MySQL ON clause to left outer-join the table
                onclause = build_onclause(db_graph, previous_table, table)
                joined_tables = join(joined_tables, table_object,
                                     isouter=True, onclause=onclause)

    return joined_tables


# SQLALCHEMY COLUMN HANDLING
# Functions to deal with column formats and constructing column expressions.
# -----------------------------------------------------------------------------
def build_where_clause(db_graph, filter_expression):
    """Creates a SQLAlchemy BinaryExpression object from a MySQL WHERE
       clause expression.

    :param db_graph: SQLAlchemy structured NetworkX Graph object.
    :type db_graph: Graph
    :param filter_expression: MySQL where clause expression.
    :type filter_expression: str
    :returns: MySQL expression-related SQLAlchemy BinaryExpression object.
    :rtype: BinaryExpression
    """
    filter_params = parsing.parse_filter(filter_expression)

    # Retrieve Table and Column objects from split filter expression strings
    table_object = db_graph.nodes[filter_params[0]]["table"]

    column_name = parsing.translate_column(
                                db_graph.graph["metadata"],
                                f"{table_object.name}.{filter_params[1]}")
    column_object = table_object.columns[column_name]

    # Checks the operator and Column type compatability
    parsing.check_operator(filter_params[2], column_object)

    right = filter_params[3]  # Stores the expressions 'right' value

    if column_object.type.python_type == bytes:
        if isinstance(right, list):
            for i in range(len(right)):
                right[i] = right[i].encode("utf-8")
        else:
            right = right.encode("utf-8")

    if right == "None":
        right = None

    where_clause = None

    if filter_params[2] == "=":
        where_clause = (column_object == right)
    elif filter_params[2] == "LIKE":
        where_clause = (column_object.like(right))
    elif filter_params[2] == "!=" or filter_params[2] == "IS NOT":
        where_clause = (column_object != right)
    elif filter_params[2] == ">":
        where_clause = (column_object > right)
    elif filter_params[2] == ">=":
        where_clause = (column_object >= right)
    elif filter_params[2] == "<":
        where_clause = (column_object < right)
    elif filter_params[2] == "<=":
        where_clause = (column_object <= right)
    elif filter_params[2] == "IN":
        where_clause = (column_object.in_(right))
    elif filter_params[2] == "NOT IN":
        where_clause = (column_object.notin_(right))

    return where_clause


def extract_column(column, check=None):
    """Get a column from a supported SQLAlchemy Column-related object.

    :param column: SQLAlchemy Column-related object.
    :type column: BinaryExpression
    :type column: Column
    :type column: count
    :type column: UnaryExpression
    :param check: SQLAlchemy Column-related object type.
    :type check: <type Column>
    :type check: <type count>
    :type check: <type UnaryExpression>
    :type check: <type BinaryExpression>
    :returns: Corresponding SQLAlchemy Column object.
    :rtype: Column
    """
    if check is not None:
        if not isinstance(column, check):
            raise TypeError(f"Type {type(column)} object passed as column "
                            "failed a specified type check.")

    if isinstance(column, Column):
        pass
    # For handling SQLAlchemy Table input
    elif isinstance(column, Table):
        column = list(column.primary_key.columns)[0]
    # For handling SQLAlchemy count(Column) expressions
    elif isinstance(column, functions.count):
        column = column.clauses.clauses[0]
    # For handling SQLAlchemy Column.distinct expressions
    elif isinstance(column, UnaryExpression):
        column = column.element
    elif isinstance(column, Label):
        column = column.element
    elif isinstance(column, Grouping):
        column = column.element.element
    # For handling SQLAlchemy comparison expressions
    elif isinstance(column, BinaryExpression):
        expression = column.left
        # For handling SQLAlchemy Column.in_() expressions
        if isinstance(expression, UnaryExpression):
            column = expression.element
        # For handling SQLAlchemy count(Column) comparisons
        elif isinstance(expression, functions.count):
            column = expression.clauses.clauses[0]
        # For handling SQLAlchemy Column.distinct() comparisons
        elif isinstance(expression, Grouping):
            column = expression.element.element
        elif isinstance(expression, Column):
            column = expression
        else:
            raise TypeError(f"BinaryExpression type {type(expression)} "
                            f"of expression {expression} is not supported.")
    elif isinstance(column, DeclarativeMeta):
        try:
            table = column.__table__
        except:
            raise ValueError("SQLAlchemy Map-related object is not supported.")
        finally:
            column = list(table.primary_key.columns)[0]
    else:
        raise TypeError(f"Input type {type(column)} is not a derivative "
                        "of a SqlAlchemy Column.")

    return column


def extract_columns(columns, check=None):
    """Get a column from a supported SQLAlchemy Column-related object(s).

    :param column: SQLAlchemy Column-related object.
    :type column: BinaryExpression
    :type column: Column
    :type column: count
    :type column: list
    :type column: UnaryExpression
    :param check: SQLAlchemy Column-related object type.
    :type check: <type Column>
    :type check: <type count>
    :type check: <type UnaryExpression>
    :type check: <type BinaryExpression>
    :returns: List of SQLAlchemy Column objects.
    :rtype: list[Column]
    """
    extracted_columns = []

    if isinstance(columns, list) or isinstance(columns, BooleanClauseList):
        for column in columns:
            if isinstance(column, BooleanClauseList):
                extracted_columns = extracted_columns + extract_columns(column)
            elif isinstance(column, Grouping):
                extracted_columns = extracted_columns + \
                                    extract_columns(column.element)
            else:
                extracted_columns.append(extract_column(column, check=check))
    elif columns is None:
        pass
    else:
        extracted_columns.append(extract_column(columns, check=check))

    return extracted_columns


# SQLALCHEMY SELECT HANDLING
# Functions that build and modify SqlAlchemy executable objects.
# -----------------------------------------------------------------------------
def build_select(db_graph, columns, where=None, order_by=None, add_in=None,
                 having=None, group_by=None):
    """Get MySQL SELECT expression SQLAlchemy executable.

    :param db_graph: SQLAlchemy structured NetworkX Graph object.
    :type db_graph: Graph
    :param columns: SQLAlchemy Column object(s).
    :type columns: Column
    :type columns: list
    :param where: MySQL WHERE clause-related SQLAlchemy object(s).
    :type where: BinaryExpression
    :type where: list
    :param order_by: MySQL ORDER BY clause-related SQLAlchemy object(s).
    :type order_by: Column
    :type order_by: list
    :param add_in: MySQL Column-related inputs to be considered for joining.
    :type add_in: Column
    :type add_in: list
    :param having: MySQL HAVING clause-related SQLAlchemy object(s).
    :type having: BinaryExpression
    :type having: list
    :param group_by: MySQL GROUP BY clause-related SQLAlchemy object(s).
    :type group_by: Column
    :type group_by: list
    :returns: MySQL SELECT expression-related SQLAlchemy executable.
    :rtype: Select
    """
    where_columns = extract_columns(where)
    having_columns = extract_columns(having)
    add_in_columns = extract_columns(add_in, check=Column)
    order_by_columns = extract_columns(order_by, check=Column)
    group_by_columns = extract_columns(group_by, check=Column)

    if not isinstance(columns, list):
        columns = [columns]

    # Gathers all the tables from the SELECT, WHERE, and ORDER BY clauses
    total_columns = (columns + where_columns + order_by_columns
                     + add_in_columns + having_columns + group_by_columns)
    fromclause = build_fromclause(db_graph, total_columns)

    select_query = select(columns).select_from(fromclause)
    select_query = append_group_by_clauses(select_query, group_by)
    select_query = append_order_by_clauses(select_query, order_by)
    select_query = append_where_clauses(select_query, where)
    select_query = append_having_clauses(select_query, having)

    return select_query


def build_count(db_graph, columns, where=None, add_in=None):
    """Get MySQL COUNT() expression SQLAlchemy executable.

    :param db_graph: SQLAlchemy structured NetworkX Graph object.
    :type db_graph: Graph
    :param columns: SQLAlchemy Column object(s).
    :type columns: Column
    :type columns: list
    :param where: MySQL WHERE clause-related SQLAlchemy object(s).
    :type where: BinaryExpression
    :type where: list
    :param add_in: MySQL Column-related inputs to be considered for joining.
    :type add_in: Column
    :type add_in: list
    :returns: MySQL COUNT() expression-related SQLAlchemy executable.
    :rtype: Select
    """
    where_columns = extract_columns(where)
    add_in_columns = extract_columns(add_in, check=Column)

    if not isinstance(columns, list):
        columns = [columns]

    # Gathers all the tables from the SELECT, WHERE, and ORDER BY clauses
    total_columns = columns + where_columns + add_in_columns
    fromclause = build_fromclause(db_graph, total_columns)
    # Converts SQLAlchemy Columns into SQLAlchemy COUNT-related objects
    column_params = []
    for column_param in columns:
        column_params.append(func.count(column_param))

    count_query = select(column_params).select_from(fromclause)
    count_query = append_where_clauses(count_query, where)

    return count_query


def build_distinct(db_graph, columns, where=None, order_by=None, add_in=None):
    """Get MySQL DISTINCT expression SQLAlchemy executable.

    :param db_graph: SQLAlchemy structured NetworkX Graph object.
    :type db_graph: Graph
    :param columns: SQLAlchemy Column object(s).
    :type columns: Column
    :type columns: list
    :param where: MySQL WHERE clause-related SQLAlchemy object(s).
    :type where: BinaryExpression
    :type where: list
    :param order_by: MySQL ORDER BY clause-related SQLAlchemy object(s).
    :type order_by: Column
    :type order_by: list
    :param add_in: MySQL Column-related inputs to be considered for joining.
    :type add_in: Column
    :type add_in: list
    :returns: MySQL DISTINCT expression-related SQLAlchemy executable.
    :rtype: Select
    """
    query = build_select(db_graph, columns, where=where,
                         order_by=order_by, add_in=add_in)

    distinct_query = query.distinct()  # Converts a SELECT to a DISTINCT
    return distinct_query


def append_where_clauses(executable, where_clauses):
    """Add WHERE SQLAlchemy BinaryExpression objects to a Select object.

    :param executable: SQLAlchemy executable query object.
    :type executable: Select
    :param where_clauses: MySQL WHERE clause-related SQLAlchemy object(s).
    :type where_clauses: BinaryExpression
    :type where_clauses: list
    :returns: MySQL expression-related SQLAlchemy exectuable.
    :rtype: Select
    """
    if where_clauses is None:
        return executable

    if isinstance(where_clauses, list):
        executable = executable.where(and_(*where_clauses))
    else:
        executable = executable.where(where_clauses)

    return executable


def append_having_clauses(executable, having_clauses):
    """Add HAVING SQLAlchemy Column objects to a Select object.

    :param executable: SQLAlchemy executable query object.
    :type executable: Select
    :param having_clauses: MySQL HAVING clause-related SQLAlchemy object(s).
    :type order_by_clauses: Column
    :type order_by_clauses: List
    :returns MySQL expression-related SQLAlchemy executable.
    :rtype: Select
    """
    if having_clauses is None:
        return executable

    if isinstance(having_clauses, list):
        executable = executable.having(and_(*having_clauses))
    else:
        executable = executable.having(having_clauses)

    return executable


def append_order_by_clauses(executable, order_by_clauses):
    """Add ORDER BY SQLAlchemy Column objects to a Select object.

    :param executable: SQLAlchemy executable query object.
    :type executable: Select
    :param order_by_clauses: MySQL ORDER BY clause-related SQLAlchemy object(s)
    :type order_by_clauses: Column
    :type order_by_clauses: list
    :returns: MySQL expression-related SQLAlchemy exectuable.
    :rtype: Select
    """
    if order_by_clauses is None:
        return executable

    if isinstance(order_by_clauses, list):
        for clause in order_by_clauses:
            executable = executable.order_by(clause)
    else:
        executable = executable.order_by(order_by_clauses)

    return executable


def append_group_by_clauses(executable, group_by_clauses):
    """Add GROUP BY SQLAlchemy Column objects to a Select object.

    :param executable: SQLAlchemy executable query object.
    :type executable: Select
    :param order_by_clauses: MySQL GROUP BY clause-related SQLAlchemy object(s)
    :type order_by_clauses: Column
    :type order_by_clauses: list
    :returns: MySQL expression-related SQLAlchemy exectuable.
    :rtype: Select
    """
    if group_by_clauses is None:
        return executable

    if isinstance(group_by_clauses, list):
        for clause in group_by_clauses:
            executable = executable.group_by(clause)
    else:
        executable = executable.group_by(group_by_clauses)

    return executable


# SQLALCHEMY EXECUTE QUERY FUNCTIONS
# Functions that execute SqlAlchemy select statements and handle outputs.
# -----------------------------------------------------------------------------
def execute(engine, executable, in_column=None, values=[], limit=8000,
            return_dict=True):
    """Use SQLAlchemy Engine to execute a MySQL query.

    :param engine: SQLAlchemy Engine object used for executing queries.
    :type engine: Engine
    :param executable: Input a executable MySQL query.
    :type executable: Select
    :type executable: str
    :param return_dict: Toggle whether execute returns dict or tuple.
    :type return_dict: Boolean
    :returns: Results from execution of given MySQL query.
    :rtype: list[dict]
    :rtype: list[tuple]
    """
    if values:
        if in_column is None:
            raise ValueError("Column input is required to condition "
                             "SQLAlchemy select for a set of values.")

        results = execute_value_subqueries(engine, executable,
                                           in_column, values,
                                           return_dict=return_dict,
                                           limit=limit)

    else:
        proxy = engine.execute(executable)

        results = proxy.fetchall()

        if return_dict:
            results_dicts = []
            for result in results:
                results_dicts.append(dict(result))

            results = results_dicts

    return results


def first_column(engine, executable, in_column=None, values=[], limit=8000):
    """Use SQLAlchemy Engine to execute and return the first column of fields.

    :param engine: SQLAlchemy Engine object used for executing queries.
    :type engine: Engine
    :param executable: Input an executable MySQL query.
    :type executable: Select
    :type executable: str
    :returns: A column for a set of MySQL values.
    :rtype: list[str]
    """
    if values:
        if in_column is None:
            raise ValueError("Column input is required to condition "
                             "SQLAlchemy select for a set of values.")

        values = first_column_value_subqueries(engine, executable,
                                               in_column, values,
                                               limit=limit)
    else:
        proxy = engine.execute(executable)
        results = proxy.fetchall()

        values = []
        for result in results:
            values.append(result[0])

    return values


def execute_value_subqueries(engine, executable, in_column, source_values,
                             return_dict=True, limit=8000):
    """Query with a conditional on a set of values using subqueries.

    :param engine: SQLAlchemy Engine object used for executing queries.
    :type engine: Engine
    :param executable: Input a executable MySQL query.
    :type executable: Select
    :type executable: str
    :param in_column: SQLAlchemy Column object.
    :type in_column: Column
    :param source_values: Values from specified MySQL column.
    :type source_values: list[str]
    :param return_dict: Toggle whether to return data as a dictionary.
    :type return_dict: Boolean
    :param limit: SQLAlchemy IN clause query length limiter.
    :type limit: int
    :returns: List of grouped data for each value constraint.
    :rtype: list
    """
    if not isinstance(in_column, Column):
        raise ValueError("Inputted column to conditional values against "
                         "is not a SqlAlchemy Column."
                         f"Object is instead type {type(in_column)}.")

    if not executable.is_derived_from(in_column.table):
        raise ValueError("Inputted column to conditional values against "
                         "must be a column from the table(s) joined in the "
                         "SQLAlchemy select.")

    values = []
    if in_column.type.python_type == bytes:
        source_values = basic.convert_to_encoded(source_values)

    chunked_values = basic.partition_list(source_values, limit)

    for value_chunk in chunked_values:
        subquery = executable.where(in_column.in_(value_chunk))

        proxy = engine.execute(subquery)
        results = proxy.fetchall()

        for result in results:
            if return_dict:
                result = dict(result)

            values.append(result)

    return values


def first_column_value_subqueries(engine, executable, in_column, source_values,
                                  limit=8000):
    """Query with a conditional on a set of values using subqueries.

    :param engine: SQLAlchemy Engine object used for executing queries.
    :type engine: Engine
    :param executable: Input a executable MySQL query.
    :type executable: Select
    :type executable: str
    :param in_column: SQLAlchemy Column object.
    :type in_column: Column
    :param source_values: Values from specified MySQL column.
    :type source_values: list[str]
    :param return_dict: Toggle whether to return data as a dictionary.
    :type return_dict: Boolean
    :param limit: SQLAlchemy IN clause query length limiter.
    :type limit: int
    :returns: Distinct values fetched from value constraints.
    :rtype: list
    """
    if not isinstance(in_column, Column):
        raise ValueError("Inputted column to conditional values against "
                         "is not a SqlAlchemy Column."
                         f"Object is instead type {type(in_column)}.")

    if not executable.is_derived_from(in_column.table):
        raise ValueError("Inputted column to conditional values against "
                         "must be a column from the table(s) joined in the "
                         "SQLAlchemy select.")

    values = []
    if in_column.type.python_type == bytes:
        source_values = basic.convert_to_encoded(source_values)

    chunked_values = basic.partition_list(source_values, limit)

    for value_chunk in chunked_values:
        subquery = executable.where(in_column.in_(value_chunk))

        proxy = engine.execute(subquery)
        results = proxy.fetchall()

        for result in results:
            values.append(result[0])

    values = list(OrderedDict.fromkeys(values))
    return values


def query(session, db_graph, table_map, where=None):
    """Use SQLAlchemy session to retrieve ORM objects from a mapped object.

    :param session: Bound and connected SQLAlchemy Session object.
    :type session: Session
    :param table_map: SQLAlchemy ORM map object.
    :param where: MySQL WHERE clause-related SQLAlchemy object(s).
    :type where: BinaryExpression
    :type where: list
    :param order_by: MySQL ORDER BY clause-related SQLAlchemy object(s).
    :type order_by: Column
    :type order_by: list
    :returns: List of mapped object instances.
    :rtype: list
    """
    where_columns = extract_columns(where)

    total_columns = [table_map] + where_columns
    from_clause = build_fromclause(db_graph, total_columns)

    query_obj = session.query(table_map).select_from(from_clause)

    if where is not None:
        if isinstance(where, list):
            query_obj = query_obj.filter(and_(*where))
        else:
            query_obj = query_obj.filter(where)

    instances = query_obj.all()
    return instances



"""Represents a collection of data about a genome that are commonly used to
maintain and update SEA-PHAGES phage genomics data.
"""
from operator import attrgetter

from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
from Bio.SeqUtils import GC

from pdm_utils.classes import evaluation, cds, trna, tmrna, source
from pdm_utils.constants import constants
from pdm_utils.functions import basic


class Genome:
    """Class to hold data about a phage genome."""

    def __init__(self):

        # The following attributes are common to any genome.
        self.nucleic_acid_type = "" # dsDNA, ssDNA, etc.
        self.order = "" # Caudovirales, Nidovirales, etc.
        self.family = "" # Siphoviridae, Myoviridae, etc.
        self.cluster = "" # A, B, C, Singleton, etc.
        self.subcluster = "" #A1, A2, etc.
        self.id = "" # Unique identifier. Case sensitive, no "_Draft".
        self.name = "" # Case sensitive and contains "_Draft".
        self.seq = Seq("", IUPAC.ambiguous_dna) # Biopython Seq object
        self.length = 0 # Size of the nucleotide sequence
        self.gc = -1 # %GC content
        self.host_genus = ""
        self.accession = ""
        self.lifestyle = "" # E.g. temperate, obligately lytic, unknown, etc.
        self.translation_table = 0

        # The following attributes are common to MySQL database.
        self.annotation_status = "" # Final, Draft, Unknown version of genome data
        self.annotation_author = -1 # 1 (can be changed), 0 (can not be changed)
        self.retrieve_record = -1 # 1 (auto update), 0 (do not auto update)
        self.date = constants.EMPTY_DATE # Used for the DateLastModified field.

        # The following attributes are common to
        # GenBank-formatted flat file records.
        self.description = ""
        self._description_name = ""
        self._description_host_genus = ""
        self.source = ""
        self._source_name = ""
        self._source_host_genus = ""
        self.organism = ""
        self._organism_name = ""
        self._organism_host_genus = ""
        self.authors = "" # Compiled string of all authors named in the record.

        # The following attributes are computed datafields that are
        # common to annotated genomes.
        self.cds_features = [] # List of all parsed CDS features
        self._cds_features_tally = 0
        self._cds_start_end_ids = []
        self._cds_end_orient_ids = []
        self._cds_descriptions_tally = 0
        self._cds_products_tally = 0
        self._cds_functions_tally = 0
        self._cds_notes_tally = 0
        self._cds_unique_start_end_ids = set() # TODO still in development.
        self._cds_duplicate_start_end_ids = set() # TODO still in development.
        self._cds_unique_end_orient_ids = set() # TODO still in development.
        self._cds_duplicate_end_orient_ids = set() # TODO still in development.

        self.trna_features = []
        self._trna_features_tally = 0
        self._trna_start_end_ids = []
        self._trna_end_orient_ids = []

        self.tmrna_features = []
        self._tmrna_features_tally = 0
        self._tmrna_start_end_ids = []
        self._tmrna_end_orient_ids = []

        self.source_features = []
        self._source_features_tally = 0

        # The following attributes are useful for processing data
        # from various data sources.
        self.filename = "" # The file name from which the data is derived
        self.type = "" # Identifier to describes how this genome is used
                       # (e.g. import, MySQL database, PhagesDB, etc.)
        self.evaluations = [] # List of warnings and errors about the data
        self.misc = None # Unstructured attribute to store misc. data not
                         # applicable to any of the other attributes and that
                         # may be used differently for downstream applications.

    def __str__(self):
        str_list = []
        if self.id != "":
            str_list.append(f"ID: {self.id}")
        if self.type != "":
            str_list.append(f"Type: {self.type}")
        if self.filename != "":
            str_list.append(f"Filename: {self.filename}")
        return ", ".join(str_list)


    def set_filename(self, filepath):
        """Set the filename. Discard the path and file extension.

        :param filepath: name of the file reference.
        :type filepath: Path
        """
        self.filename = filepath.stem
        # split_filepath = value.split('/')
        # filename = split_filepath[-1]
        # filename = filename.split('.')[0]
        # self.filename = filename

    def set_id(self, value=None, attribute=None):
        """Set the id from either an input value or an indicated attribute.

        :param value:
            unique identifier for the genome.
        :type value: str
        :param attribute:
            name of a genome object attribute that stores
            a unique identifier for the genome.
        :type attribute: str
        """

        if value is None:
            if (attribute is not None and hasattr(self, attribute)):
                value = getattr(self, attribute)
            else:
                value = ""
        self.id = basic.edit_suffix(value, "remove")


    def set_host_genus(self, value=None, attribute=None, format="empty_string"):
        """Set the host_genus from a value parsed from the indicated attribute.

        The input data is split into multiple parts, and the first word is
        used to set host_genus.

        :param value:
            the host genus of the phage genome
        :type value: str
        :param attribute:
            the name of the genome attribute from which the
            host_genus attribute will be set
        :type attribute: str
        :param format:
            the default format if the input is an empty/null value.
        :type format: str
        """

        if value is None:
            if (attribute is not None and hasattr(self, attribute)):
                value = getattr(self, attribute)
            else:
                value = ""
        if isinstance(value, str):
            value = value.strip()

        # The host_genus value may need to be split. But don't split until
        # it is determined if the value is a null value.
        value = basic.convert_empty(value, "empty_string")
        if value != "":
            self.host_genus = value.split(" ")[0]
        else:
            self.host_genus = basic.convert_empty(value, format)


    def parse_description(self):
        """Retrieve the name and host_genus from the 'description' attribute."""
        string = self.description
        name, host_genus = \
            basic.parse_names_from_record_field(string)
        self._description_name = name
        self._description_host_genus = host_genus


    def parse_source(self):
        """Retrieve the name and host_genus from the 'source' attribute."""
        string = self.source
        name, host_genus = \
            basic.parse_names_from_record_field(string)
        self._source_name = name
        self._source_host_genus = host_genus


    def parse_organism(self):
        """Retrieve the name and host_genus from the 'organism' attribute."""
        string = self.organism
        name, host_genus = \
            basic.parse_names_from_record_field(string)
        self._organism_name = name
        self._organism_host_genus = host_genus


    def set_sequence(self, value):
        """Set the nucleotide sequence and compute the length.

        This method coerces sequences into a Biopython Seq object.

        :param value: the genome's nucleotide sequence.
        :type value: str or Seq
        """
        if not isinstance(value, Seq):
            self.seq = Seq(value).upper()
        else:
            self.seq = value.upper()
        self.length = len(self.seq)
        if self.length > 0:
            self.gc = round(GC(self.seq), 4)
        else:
            self.gc = -1


    def set_accession(self, value, format="empty_string"):
        """Set the accession.

        The Accession field in the MySQL database defaults to ''.
        Some flat file accessions have the version number suffix, so discard
        the version number.

        :param value:
            GenBank accession number.
        :type value: str
        :param format:
            indicates the format of the data if it is not a
            valid accession. Default is ''.
        :type format: misc.
        """
        if isinstance(value, str):
            value = value.strip()
            value = value.split(".")[0]
        self.accession = basic.convert_empty(value, format)

    def set_cds_features(self, value):
        """Set and tally the CDS features.

        :param value: list of Cds objects.
        :type value: list
        """
        self.cds_features = value # Should be a list.
        self._cds_features_tally = len(self.cds_features)

    def set_cds_id_list(self):
        """Creates lists of CDS feature identifiers.

        The first identifier is derived from the start and end coordinates.
        The second identifier is derived from the transcription end
        coordinate and orientation.
        """
        start_end_id_list = []
        end_orient_id_list = []
        for cds_ftr in self.cds_features:
            start_end_id_list.append(cds_ftr._start_end_id)
            end_orient_id_list.append(cds_ftr._end_orient_id)
        self._cds_start_end_ids = start_end_id_list
        self._cds_end_orient_ids = end_orient_id_list

    def set_trna_features(self, value):
        """Set and tally the tRNA features.

        :param value: list of Trna objects.
        :type value: list
        """
        self.trna_features = value      # Should be a list
        self._trna_features_tally = len(self.trna_features)

    def set_tmrna_features(self, value):
        """Set and tally the tmRNA features.
        :param value: list of Tmrna objects.
        :type value: list
        """
        self.tmrna_features = value     # Should be a list
        self._tmrna_features_tally = len(self.tmrna_features)

    def set_source_features(self, value):
        """Set and tally the source features.

        :param value: list of Source objects.
        :type value: list
        """
        self.source_features = value # Should be a list
        self._source_features_tally = len(self.source_features)


    def set_cluster(self, value):
        """Set the cluster and modify singleton if needed.

        :param value: Cluster designation of the genome.
        :type value: str
        """
        singleton = "Singleton"
        if isinstance(value, str):
            value = value.strip()

            # PhagesDB-output format.
            if value.capitalize() == singleton:
                self.cluster = singleton
            else:
                self.cluster = value

        # MySQL database-output format
        if value is None:
            self.cluster = singleton


    def set_subcluster(self, value):
        """Set the subcluster.

        :param value: Subcluster designation of the genome.
        :type value: str
        """
        if isinstance(value, str):
            self.subcluster = value.strip()

        # MySQL database-output format
        if value is None:
            self.subcluster = "none"


    def set_date(self, value, format="empty_datetime_obj"):
        """Set the date attribute.

        :param value: Date
        :type value: misc
        :param format: Indicates the format if the value is empty.
        :type format: str
        """
        self.date = basic.convert_empty(value, format)


    def set_annotation_author(self, value):
        """Convert annotation_author to integer value if possible.

        :param value: Numeric value.
        :type value: str, int
        """
        try:
            self.annotation_author = int(value)
        except:
            self.annotation_author = value


    def set_retrieve_record(self, value):
        """Convert retrieve_record to integer value if possible.

        :param value: Numeric value.
        :type value: str, int
        """
        try:
            self.retrieve_record = int(value)
        except:
            self.retrieve_record = value


    def set_cds_descriptions(self, value):
        """Set each CDS processed description as indicated.

        :param value: Name of the description field.
        :type value: str
        """
        for x in range(len(self.cds_features)):
            cds_ftr = self.cds_features[x]
            cds_ftr.set_description(value)


    def tally_cds_descriptions(self):
        """Tally the non-generic CDS descriptions."""
        self._cds_descriptions_tally = 0
        self._cds_products_tally = 0
        self._cds_functions_tally = 0
        self._cds_notes_tally = 0
        for cds_ftr in self.cds_features:
            if cds_ftr.description != "":
                self._cds_descriptions_tally += 1
            if cds_ftr.product != "":
                self._cds_products_tally += 1
            if cds_ftr.function != "":
                self._cds_functions_tally += 1
            if cds_ftr.note != "":
                self._cds_notes_tally += 1


    def set_unique_cds_start_end_ids(self):
        """Identify CDS features contain unique start-end coordinates."""
        unique_id_tuples, duplicate_id_tuples = \
            basic.identify_unique_items(self._cds_start_end_ids)
        self._cds_unique_start_end_ids = set(unique_id_tuples)
        self._cds_duplicate_start_end_ids = set(duplicate_id_tuples)


    def set_unique_cds_end_orient_ids(self):
        """Identify CDS features contain unique transcription end-orientation coordinates."""
        unique_id_tuples, duplicate_id_tuples = \
            basic.identify_unique_items(self._cds_end_orient_ids)
        self._cds_unique_end_orient_ids = set(unique_id_tuples)
        self._cds_duplicate_end_orient_ids = set(duplicate_id_tuples)


    def set_feature_ids(self, use_type=False, use_cds=False,
                        use_trna=False, use_tmrna=False, use_source=False):
        """Sets the id of each feature.

        Lists of features can be added to this method. The method assumes
        that all elements in all lists contain 'id', 'start', and 'stop'
        attributes. This feature attribute is processed within
        the Genome object because and not within the feature itself since
        the method sorts all features and generates systematic IDs based on
        feature order in the genome.

        :param use_type:
            Indicates whether the type of object should be
            added to the feature id.
        :type use_type: bool
        :param use_cds:
            Indicates whether ids for CDS features should be generated.
        :type use_cds: bool
        :param use_trna:
            Indicates whether ids for tRNA features should be generated.
        :type use_trna: bool
        :param use_tmrna:
            Indicates whether ids for tmRNA features should be generated.
        :type use_tmrna: bool
        :param use_source:
            Indicates whether ids for source features should be generated.
        :type use_source: bool
        """

        # Both coordinates are used to control the order of features
        # that may share one, but not both, coordinates (e.g. tail
        # assembly chaperone).
        list_to_sort = []
        if use_cds:
            list_to_sort.extend(self.cds_features)
        if use_source:
            list_to_sort.extend(self.source_features)
        if use_trna:
            list_to_sort.extend(self.trna_features)
        if use_tmrna:
            list_to_sort.extend(self.tmrna_features)

        sorted_list = sorted(list_to_sort, key=attrgetter("start", "stop"))

        for index in range(len(sorted_list)):
            if use_type:
                if isinstance(sorted_list[index], cds.Cds):
                    delimiter = "_CDS_"
                elif isinstance(sorted_list[index], source.Source):
                    delimiter = "_SRC_"
                elif isinstance(sorted_list[index], trna.Trna):
                    delimiter = "_TRNA_"
                elif isinstance(sorted_list[index], tmrna.Tmrna):
                    delimiter = "_TMRNA_"
                else:
                    delimiter = "_"
            else:
                delimiter = "_"

            sorted_list[index].id = self.id + delimiter + str(index + 1)


    def update_name_and_id(self, value):
        """Update the genome name and id in all locations in a Genome object.

        :param gnm: A pdm_utils Genome object.
        :type gnm: Genome
        :param value: Value used to update the Genome id and name.
        :type value: str
        """
        # Treat the new value as a name, which could have '_Draft' suffix,
        # then use set_id() to remove suffix.
        self.name = value
        self.set_id(value=value)
        self.set_feature_ids(use_type=True, use_cds=True)
        self.set_feature_ids(use_type=True, use_source=True)
        self.set_feature_ids(use_type=True, use_trna=True)
        self.set_feature_ids(use_type=True, use_tmrna=True)
        self.set_feature_genome_ids(use_cds=True, use_trna=True,
                                      use_tmrna=True, use_source=True)


    # TODO add parameters to specify which feature types
    # (e.g. cds=True, trna=True, ...).
    def clear_locus_tags(self):
        """Resets locus_tags to empty string."""
        for x in range(len(self.cds_features)):
            self.cds_features[x].locus_tag = ""
        for x in range(len(self.trna_features)):
            self.trna_features[x].locus_tag = ""
        for x in range(len(self.tmrna_features)):
            self.tmrna_features[x].locus_tag = ""

    def set_feature_genome_ids(self, use_cds=False, use_trna=False,
                               use_tmrna=False, use_source=False, value=None):
        """Sets the genome_id of each feature.

        :param use_cds:
            Indicates whether genome_id for CDS features should be set.
        :type use_cds: bool
        :param use_trna:
            Indicates whether genome_id for tRNA features should be set.
        :type use_trna: bool
        :param use_tmrna:
            Indicates whether genome_id for tmRNA features should be set.
        :type use_tmrna: bool
        :param use_source:
            Indicates whether genome_id for source features should be set.
        :type use_source: bool
        :param value: Genome identifier.
        :type value: str
        """
        if value is None:
            value = self.id

        feature_list = []
        if use_cds == True:
            feature_list.extend(self.cds_features)
        if use_source == True:
            feature_list.extend(self.source_features)
        if use_trna == True:
            feature_list.extend(self.trna_features)
        if use_tmrna == True:
            feature_list.extend(self.tmrna_features)

        for feature in feature_list:
            feature.genome_id = value


    # Evaluations.
    def set_eval(self, eval_id, definition, result, status):
        """Constructs and adds an Evaluation object to the evaluations list.

        :param eval_id: Unique identifier for the evaluation.
        :type eval_id: str
        :param definition: Description of the evaluation.
        :type definition: str
        :param result: Description of the outcome of the evaluation.
        :type result: str
        :param status: Outcome of the evaluation.
        :type status: str
        """
        evl = evaluation.Evaluation(eval_id, definition, result, status)
        self.evaluations.append(evl)


    def check_attribute(self, attribute, check_set, expect=False, eval_id=None,
                        success="correct", fail="error", eval_def=None):
        """Check that the attribute value is valid.

        :param attribute: Name of the Genome object attribute to evaluate.
        :type attribute: str
        :param check_set: Set of reference ids.
        :type check_set: set
        :param expect:
            Indicates whether the attribute value is expected to be present
            in the check set.
        :type expect: bool
        :param eval_id: Unique identifier for the evaluation.
        :type eval_id: str
        :param success: Default status if the outcome is a success.
        :type success: str
        :param fail: Default status if the outcome is not a success.
        :type fail: str
        :param eval_def: Description of the evaluation.
        :type eval_def: str
        """
        try:
            test = True
            value1 = getattr(self, attribute)
        except:
            test = False
            value1 = None
        if test:
            value1_short = basic.truncate_value(str(value1), 30, "...")
            result = f"The {attribute} value '{value1_short}' is "

            value2 = basic.check_value_expected_in_set(
                        value1, check_set, expect)
            if value2:
                result = result + "valid."
                status = success
            else:
                result = result + "not valid."
                status = fail
        else:
            result = f"'{attribute}' is not a valid attribute to be evaluated."
            status = "untested"
        definition = f"Check the value of the '{attribute}' attribute."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def compare_two_attributes(self, attribute1, attribute2,
                               expect_same=False, eval_id=None,
                               success="correct", fail="error", eval_def=None):
        """Determine if two attributes are the same.

        :param attribute1: First attribute to compare.
        :type attribute1: str
        :param attribute2: Second attribute to compare.
        :type attribute2: str
        :param expect_same:
            Indicates whether the two attribute values are expected to be
            the same.
        :type expect_same: bool
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        try:
            test = True
            value1 = getattr(self, attribute1)
            value2 = getattr(self, attribute2)
        except:
            test = False
            value1 = None
            value2 = None
        if test:
            if value1 == value2:
                actual_same = True
            else:
                actual_same = False

            v1_short = basic.truncate_value(str(value1), 30, "...")
            v2_short = basic.truncate_value(str(value2), 30, "...")
            result = (f"The '{attribute1}' attribute contains: '{v1_short}'. "
                      f"The '{attribute2}' attribute contains: '{v2_short}'. "
                      "These two values are ")

            if actual_same:
                result = result + "identical, "
            else:
                result = result + "different, "

            if actual_same and expect_same:
                result = result + "as expected."
                status = success
            elif not actual_same and not expect_same:
                result = result + "as expected."
                status = success
            else:
                result = result + "which is not expected."
                status = fail
        else:
            result = (f"'{attribute1}' and/or '{attribute2}' is "
                      "not a valid field to be compared.")
            status = "untested"
        definition = f"Compare values of '{attribute1}' and '{attribute2}' attributes."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_cluster_structure(self, eval_id=None, success="correct",
                                fail="error", eval_def=None):
        """Check whether the cluster attribute is structured appropriately.

        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        result = f"The Cluster is '{self.cluster}'. "
        if self.cluster != "none":
            result = result + "It is structured "
            left, right = basic.split_string(self.cluster)

            if (right != "" or left.isalpha() == False):
                result = result + "incorrectly."
                status = fail
            else:
                result = result + "correctly."
                status = success
        else:
            result = result + "It is empty."
            status = "untested"
        definition = "Check if the Cluster attribute is structured correctly."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_subcluster_structure(self, eval_id=None, success="correct",
                                   fail="error", eval_def=None):
        """Check whether the subcluster attribute is structured appropriately.

        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        result = f"The Subcluster is '{self.subcluster}'. "
        if (self.subcluster != "none" and self.subcluster != ""):
            result = result + "It is structured "
            left, right = basic.split_string(self.subcluster)
            if (left.isalpha() == False or right.isdigit() == False):
                result = result + "incorrectly."
                status = fail
            else:
                result = result + "correctly."
                status = success
        else:
            result = result + "It is empty."
            status = "untested"
        definition = "Check if the Subcluster attribute is structured correctly."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_compatible_cluster_and_subcluster(self, eval_id=None,
                        success="correct", fail="error", eval_def=None):
        """Check compatibility of cluster and subcluster attributes.

        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        result = (f"The Cluster is '{self.cluster}', "
                  f"the Subcluster is '{self.subcluster}', and they are ")
        output = basic.compare_cluster_subcluster(self.cluster, self.subcluster)
        if not output:
            result = result + "not compatible."
            status = fail
        else:
            result = result + "compatible."
            status = success
        definition = "Check for compatibility between Cluster and Subcluster."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_nucleotides(self, check_set=set(), eval_id=None,
                          success="correct", fail="error", eval_def=None):
        """Check if all nucleotides in the sequence are expected.

        :param check_set: Set of reference nucleotides.
        :type check_set: set
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        # When Biopython SeqIO parses the GenBank record, it automatically
        # determines that it is a DNA sequence. It assigns the Seq object
        # alphabet as IUPACAmbiguousDNA. The alphabet could be coerced
        # to a different alphabet, and then tested using the
        # Bio.Alphabet._verify_alphabet() function. Since this is a private
        # function though, it is not clear how stable/reliable it is.
        # Instead, Bio.Alphabet.IUPAC.unambiguous_dna alphabet can be passed
        # to the check_nucleotides method.
        nucleotide_set = set(self.seq)
        nucleotide_error_set = nucleotide_set - check_set
        if len(nucleotide_error_set) > 0:
            nes_string = basic.join_strings(nucleotide_error_set, delimiter=", ")
            result = ("There are unexpected nucleotides in the sequence: "
                      f"{nes_string}.")
            status = fail
        else:
            result = "There are no unexpected nucleotides in the sequence."
            status = success
        definition = "Check if all nucleotides in the sequence are expected."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_authors(self, check_set=set(), expect=True, eval_id=None,
                      success="correct", fail="error", eval_def=None):
        """Check author list.

        Evaluates whether at least one author in the in the list of
        authors is present in a set of reference authors.

        :param check_set: Set of reference authors.
        :type check_set: set
        :param expect:
            Indicates whether at least one author in the
            list of authors is expected to be present in the check set.
        :type expect: bool
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        authors = self.authors.lower()
        authors = authors.replace(";", ",")
        authors = authors.replace(" ", ",")
        authors_list = authors.split(",")
        authors_list = [x.strip() for x in authors_list]
        authors_set = set(authors_list)
        mutual_authors_set = authors_set & check_set
        if len(mutual_authors_set) == 0:
            missing_set = check_set - authors_set
            missing_string = basic.join_strings(missing_set, delimiter=", ")
            result = ("The following authors are not "
                      f"listed: {missing_string}. This is ")
            if expect:
                result = result + "not expected."
                status = fail
            else:
                result = result + "expected."
                status = success
        else:
            mas_string = basic.join_strings(mutual_authors_set, delimiter=", ")
            result = f"The following authors are listed: {mas_string}. This is "
            if expect:
                result = result + "expected."
                status = success
            else:
                result = result + "not expected."
                status = fail
        definition = "Check authorship."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_magnitude(self, attribute, expect, ref_value, eval_id=None,
                        success="correct", fail="error", eval_def=None):
        """Check that the magnitude of a numerical attribute is valid.

        :param attribute: same as for check_attribute().
        :param expect:
            Comparison symbol indicating direction of magnitude (>, =, <).
        :type expect: str
        :param ref_value: Numerical value for comparison.
        :type ref_value: int, float, datetime
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        try:
            test = True
            query_value = getattr(self, attribute)
        except:
            test = False
            query_value = None
        if test:
            result = f"The {attribute} value {query_value} is "
            if query_value > ref_value:
                compare = ">"
                result = result + "greater than "
            elif query_value == ref_value:
                compare = "="
                result = result + "equal to "
            else:
                compare = "<"
                result = result + "less than "
            result = result + f"{ref_value}, which is "
            if compare == expect:
                result = result + "expected."
                status = success
            else:
                result = result + "not expected."
                status = fail
        else:
            result = f"'{attribute}' is not a valid attribute to be evaluated."
            status = "untested"
        definition = f"Check the magnitude of the '{attribute}' attribute."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    # TODO this may no longer be needed.
    def check_cds_start_end_ids(self, eval_id=None, success="correct",
                                fail="error", eval_def=None):
        """Check if there are any duplicate start-end coordinates.

        Duplicated start-end coordinates may represent
        unintentional duplicate CDS features.

        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """

        if len(self._cds_duplicate_start_end_ids) > 0:
            result = ("There are multiple CDS features with the same "
                      "start and end coordinates.")
            status = fail
        else:
            result = ("All CDS features contain unique start and "
                      "end coordinate information.")
            status = success
        definition = ("Check whether CDS features can be uniquely "
                      "identified by their start and end coordinates.")
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    # TODO this may no longer be needed.
    def check_cds_end_orient_ids(self, eval_id=None, success="correct",
                                 fail="error", eval_def=None):
        """Check if there are any duplicate transcription end-orientation coordinates.

        Duplicated transcription end-orientation coordinates may represent
        unintentional duplicate CDS features with slightly
        different start coordinates.

        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """

        if len(self._cds_duplicate_end_orient_ids) > 0:
            result = ("There are multiple CDS features with the same "
                      "transcription end coordinate and orientation.")
            status = fail
        else:
            result = ("All CDS features contain unique orientation and "
                      "transcription end coordinate information.")
            status = success
        definition = ("Check whether CDS features can be uniquely "
                      "identified by their orientation and transcription end "
                      "coordinate.")
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_feature_coordinates(self, use_cds=False, use_trna=False,
                use_tmrna=False, other=None, strand=False, eval_id=None,
                success="correct", fail="error", eval_def=None):
        """Identify nested, duplicated, or partially-duplicated
        features.

        :param use_cds:
            Indicates whether ids for CDS features should be generated.
        :type use_cds: bool
        :param use_trna:
            Indicates whether ids for tRNA features should be generated.
        :type use_trna: bool
        :param use_tmrna:
            Indicates whether ids for tmRNA features should be generated.
        :type use_tmrna: bool
        :param other: List of features that should be included.
        :type other: list
        :param strand: Indicates if feature orientation should be included.
        :type strand: bool
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        unsorted_feature_lists = []
        unsorted_features = []
        ftr_types = set()
        if use_cds:
            ftr_types.add("cds")
            unsorted_features.extend(self.cds_features)
        if use_trna:
            ftr_types.add("trna")
            unsorted_features.extend(self.trna_features)
        if use_tmrna:
            ftr_types.add("tmrna")
            unsorted_features.extend(self.tmrna_features)
        if other is not None:
            ftr_types.add("other")
            unsorted_features.extend(other)
        if strand:
            s_info = "were"
            unsorted_f_features = [] # Forward orientation
            unsorted_r_features = [] # Reverse orientation
            for index in range(len(unsorted_features)):
                feature = unsorted_features[index]
                strand = basic.reformat_strand(feature.orientation,
                                               format="fr_short")
                if strand == "f":
                    unsorted_f_features.append(feature)
                else:
                    unsorted_r_features.append(feature)
            unsorted_feature_lists.append(unsorted_f_features)
            unsorted_feature_lists.append(unsorted_r_features)
        else:
            s_info = "were not"
            unsorted_feature_lists.append(unsorted_features)
        ft_string = basic.join_strings(ftr_types, delimiter=", ")
        result = (f"The following types of features were evaluated: {ft_string}. "
                  f"Features {ft_string} separately grouped "
                  "by orientation for evaluation. ")
        msgs = ["There are one or more errors with the feature coordinates."]
        for unsorted_features in unsorted_feature_lists:
            sorted_features = sorted(unsorted_features,
                                     key=attrgetter("start", "stop"))
            index = 0
            while index < len(sorted_features) - 1:
                current = sorted_features[index]
                next = sorted_features[index + 1]
                ftrs = (f"Feature1 ID: {current.id}, "
                        f"start coordinate: {current.start}, "
                        f"stop coordinate: {current.stop}, "
                        f"orientation: {current.orientation}. "
                        f"Feature2 ID: {next.id}, "
                        f"start coordinate: {next.start}, "
                        f"stop coordinate: {next.stop}, "
                        f"orientation: {next.orientation}. ")

                if (current.start == next.start and current.stop == next.stop):
                    msgs.append(ftrs)
                    msgs.append("Feature1 and Feature2 contain identical "
                                "start and stop coordinates.")

                # To identify nested features, the following tests
                # avoid false errors due to genes that may wrap around the
                # genome.
                elif (current.start < next.start and
                      current.start < next.stop and
                      current.stop > next.start and
                      current.stop > next.stop):
                    msgs.append(ftrs)
                    msgs.append("Feature2 is nested within Feature1.")
                elif (current.start == next.start and
                        basic.reformat_strand(current.orientation,
                            format="fr_short") == "r" and
                        basic.reformat_strand(next.orientation,
                            format="fr_short") == "r"):
                    msgs.append(ftrs)
                    msgs.append(("Feature1 and Feature2 contain "
                                 "identical stop coordinates."))
                elif (current.stop == next.stop and
                        basic.reformat_strand(current.orientation,
                            format="fr_short") == "f" and
                        basic.reformat_strand(next.orientation,
                            format="fr_short") == "f"):
                    msgs.append(ftrs)
                    msgs.append(("Feature1 and Feature2 contain "
                                 "identical stop coordinates."))
                else:
                    pass
                index += 1
        if len(msgs) > 1:
            result = result + " ".join(msgs)
            status = fail
        else:
            result = result + "The feature coordinates are correct."
            status = success
        definition = ("Check if there are any feature coordinate conflicts.")
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)



"""Misc. functions to interact with NCBI databases."""

from Bio import Entrez, SeqIO

from pdm_utils.functions import basic

# GLOBAL VARIABLES
# ----------------------------------------------------------------------------
RETTYPE_MAPPINGS = {"gb": "gb", "tbl": "ft"}


# TODO unittest.
def set_entrez_credentials(tool=None, email=None, api_key=None):
    """Set BioPython Entrez credentials to improve speed and reliability.

    :param tool: Name of the software/tool being used.
    :type tool: str
    :param email: Email contact information for NCBI.
    :type email: str
    :param api_key: Unique NCBI-issued identifier to enhance retrieval speed.
    :type api_key: str
    """
    if tool is not None:
        Entrez.tool = tool
    if email is not None:
        Entrez.email = email
    if api_key is not None:
        Entrez.api_key = api_key


# TODO unittest.
def run_esearch(db="", term="", usehistory=""):
    """Search for valid records in NCBI.

    Uses NCBI esearch implemented through BioPython Entrez.

    :param db: Name of the database to search.
    :type db: str
    :param term: Search term.
    :type term: str
    :param usehistory: Indicates if prior searches should be used.
    :type usehistory: str
    :return: Results of the search for each valid record.
    :rtype: dict
    """
    search_handle = Entrez.esearch(db=db, term=term, usehistory=usehistory)
    search_record = Entrez.read(search_handle)
    search_handle.close()
    return search_record


# TODO unittest.
def get_summaries(db="", query_key="", webenv=""):
    """Retrieve record summaries from NCBI for a list of accessions.

    Uses NCBI esummary implemented through BioPython Entrez.

    :param db: Name of the database to get summaries from.
    :type db: str
    :param query_key:
        Identifier for the search.
        This can be directly generated from run_esearch().
    :type query_key: str
    :param webenv: Identifier that can be directly generated from run_esearch()
    :type webenv: str
    :return: List of dictionaries, where each dictionary is a record summary.
    :rtype: list
    """
    summary_handle = Entrez.esummary(db=db, query_key=query_key, webenv=webenv)
    summary_records = Entrez.read(summary_handle)
    summary_handle.close()
    return summary_records


# TODO test.
def get_accessions_to_retrieve(summary_records):
    """Extract accessions from summary records.

    :param summary_records:
        List of dictionaries, where each dictionary is a record summary.
    :type summary_records: list
    :return: List of accessions.
    :rtype: list
    """
    accessions = []
    for doc_sum in summary_records:
        doc_sum_accession = doc_sum["Caption"]
        accessions.append(doc_sum_accession)
    return accessions


def get_data_handle(accession_list, db="nucleotide", rettype="gb",
                    retmode="text"):
    fetch_query = ",".join(accession_list)
    fetch_handle = Entrez.efetch(db=db, id=fetch_query, rettype=rettype,
                                 retmode=retmode)

    return fetch_handle


# TODO Owen test
def get_verified_data_handle(acc_id_dict, ncbi_cred_dict={}, batch_size=200,
                             file_type="gb"):
    """Retrieve genomes from GenBank.

    output_folder = Path to where files will be saved.
    acc_id_dict = Dictionary where key = Accession and value = List[PhageIDs]
    """

    # More setup variables if NCBI updates are desired. NCBI Bookshelf resource
    # "The E-utilities In-Depth: Parameters, Syntax and More", by Dr. Eric
    # Sayers, recommends that a single request not contain more than about 200
    # UIDS so we will use that as our batch size, and all Entrez requests must
    # include the user's email address and tool name.
    set_entrez_credentials(
        tool=ncbi_cred_dict.get("tool"),
        email=ncbi_cred_dict.get("email"),
        api_key=ncbi_cred_dict.get("api_key"))

    # Use esearch to verify the accessions are valid and efetch to retrieve
    # the record
    # Create batches of accessions
    unique_accession_list = list(acc_id_dict.keys())

    # Add [ACCN] field to each accession number
    appended_accessions = \
        [accession + "[ACCN]" for accession in unique_accession_list]

    # When retrieving in batch sizes, first create the list of values
    # indicating which indices of the unique_accession_list should be used
    # to create each batch.
    # For instace, if there are five accessions, batch size of two produces
    # indices = 0,2,4

    chunked_accessions = basic.partition_list(appended_accessions, batch_size)
    for chunk in chunked_accessions:
        delimiter = " | "
        esearch_term = delimiter.join(chunk)

        # Use esearch for each accession
        search_record = run_esearch(db="nucleotide", term=esearch_term,
                                    usehistory="y")
        search_count = int(search_record["Count"])
        search_webenv = search_record["WebEnv"]
        search_query_key = search_record["QueryKey"]
        summary_records = get_summaries(db="nucleotide",
                                        query_key=search_query_key,
                                        webenv=search_webenv)

        accessions_to_retrieve = get_accessions_to_retrieve(summary_records)
        if len(accessions_to_retrieve) > 0:
            fetch_handle = get_data_handle(accessions_to_retrieve,
                                           rettype=RETTYPE_MAPPINGS[file_type])
            return fetch_handle
        else:
            return None


# TODO unittest.
def get_records(accession_list, db="nucleotide", rettype="gb", retmode="text"):
    """Retrieve records from NCBI from a list of active accessions.

    Uses NCBI efetch implemented through BioPython Entrez.

    :param accession_list: List of NCBI accessions.
    :type accession_list: list
    :param db: Name of the database to get summaries from (e.g. 'nucleotide').
    :type db: str
    :param rettype: Type of record to retrieve (e.g. 'gb').
    :type rettype: str
    :param retmode: Format of data to retrieve (e.g. 'text').
    :type retmode: str
    :return: List of BioPython SeqRecords generated from GenBank records.
    :rtype: list
    """
    retrieved_records = []
    fetch_query = ",".join(accession_list)
    fetch_handle = Entrez.efetch(db=db, id=fetch_query, rettype=rettype,
                                 retmode=retmode)
    fetch_records = SeqIO.parse(fetch_handle, "genbank")
    for record in fetch_records:
        retrieved_records.append(record)
    fetch_handle.close()
    return retrieved_records



"""Functions that are used in the phameration pipeline"""

import random
import colorsys

from pdm_utils.functions import mysqldb
from pdm_utils.functions import mysqldb_basic


# DATABASE FUNCTIONS
def get_pham_geneids(engine):
    """
    Queries the database for those genes that are already phamerated.
    :param engine: the Engine allowing access to the database
    :return: pham_geneids
    """
    pham_geneids = dict()

    geneid_query = "SELECT GeneID, PhamID FROM gene WHERE PhamID IS NOT NULL"
    geneid_results = mysqldb_basic.query_dict_list(engine, geneid_query)

    for dictionary in geneid_results:
        pham_id = dictionary["PhamID"]
        geneid = dictionary["GeneID"]

        if pham_id in pham_geneids.keys():
            pham_geneids[pham_id] = pham_geneids[pham_id] | {geneid}
        else:
            pham_geneids[pham_id] = {geneid}

    return pham_geneids


def get_pham_colors(engine):
    """
    Queries the database for the colors of existing phams
    :param engine: the Engine allowing access to the database
    :return: pham_colors
    """
    pham_colors = dict()

    color_query = "SELECT PhamID, Color FROM pham"
    color_results = mysqldb_basic.query_dict_list(engine, color_query)

    for dictionary in color_results:
        pham_id = dictionary["PhamID"]
        color = dictionary["Color"]

        pham_colors[pham_id] = color.upper()

    return pham_colors


def get_new_geneids(engine):
    """
    Queries the database for those genes that are not yet phamerated.
    :param engine: the Engine allowing access to the database
    :return: new_geneids
    """
    new_geneids = list()

    gene_query = "SELECT GeneID FROM gene WHERE PhamID IS NULL"
    gene_results = mysqldb_basic.query_dict_list(engine, gene_query)

    # At scale, much cheaper to convert list to set than to build the
    # set one gene at a time
    for dictionary in gene_results:
        geneid = dictionary["GeneID"]
        new_geneids.append(geneid)

    return set(new_geneids)


def get_geneids_and_translations(engine):
    """
    Constructs a dictionary mapping all geneids to their translations.
    :param engine: the Engine allowing access to the database
    :return: gs_to_ts
    """
    gs_to_ts = dict()

    query = ("SELECT GeneID, CONVERT(Translation USING utf8) as Translation "
             "FROM gene")
    results = mysqldb_basic.query_dict_list(engine, query)

    for dictionary in results:
        geneid = dictionary["GeneID"]
        translation = dictionary["Translation"]
        gs_to_ts[geneid] = translation

    return gs_to_ts


def get_translation_groups(engine):
    """
    Constructs a dictionary mapping all unique translations to their
    groups of geneids that share them
    :param engine: the Engine allowing access to the database
    :return: ts_to_gs
    """
    ts_to_gs = dict()

    query = ("SELECT GeneID, CONVERT(Translation USING utf8) as Translation "
             "FROM gene")
    results = mysqldb_basic.query_dict_list(engine, query)

    for dictionary in results:
        geneid = dictionary["GeneID"]
        trans = dictionary["Translation"]
        geneids = ts_to_gs.get(trans, [])
        geneids.append(geneid)
        ts_to_gs[trans] = geneids

    return ts_to_gs


def update_pham_table(colors, engine):
    """
    Populates the pham table with the new PhamIDs and their colors.
    :param colors: new pham color data
    :type colors: dict
    :param engine: sqlalchemy Engine allowing access to the database
    :return:
    """
    # First command needs to clear the pham table
    commands = ["DELETE FROM pham"]

    # Then we need to issue insert commands for each pham
    for key in colors.keys():
        commands.append(f"INSERT INTO pham (PhamID, Color) VALUES ({key}, "
                        f"'{colors[key]}')")

    mysqldb.execute_transaction(engine, commands)


def update_gene_table(phams, engine):
    """
    Updates the gene table with new pham data
    :param phams: new pham gene data
    :type phams: dict
    :param engine: sqlalchemy Engine allowing access to the database
    :return:
    """
    commands = []

    # We need to issue an update command for each gene in each pham
    for key in phams.keys():
        for gene in phams[key]:
            commands.append(f"UPDATE gene SET PhamID = {key} WHERE GeneID = '{gene}'")

    mysqldb.execute_transaction(engine, commands)


def fix_white_phams(engine):
    """
    Find any phams with 2+ members which are colored as though they are
    orphams (#FFFFFF in pham.Color).
    :param engine: sqlalchemy Engine allowing access to the database
    :return:
    """
    query = "SELECT c.PhamID FROM (SELECT g.PhamID, COUNT(GeneID) AS count, "\
            "p.Color FROM gene AS g INNER JOIN pham AS p ON g.PhamID " \
            "= p.PhamID GROUP BY PhamID) AS c WHERE Color = '#FFFFFF' "\
            "AND count > 1"

    results = mysqldb_basic.query_dict_list(engine, query)
    print(f"Found {len(results)} white phams...")

    commands = []
    for dictionary in results:
        pham_id = dictionary["PhamID"]
        h = s = v = 0
        while h <= 0:
            h = random.random()
        while s < 0.5:
            s = random.random()
        while v < 0.8:
            v = random.random()
        rgb = colorsys.hsv_to_rgb(h, s, v)
        rgb = (rgb[0] * 255, rgb[1] * 255, rgb[2] * 255)
        hexrgb = "#{:02x}{:02x}{:02x}".format(int(rgb[0]), int(rgb[1]),
                                              int(rgb[2]))
        new_color = hexrgb.upper()
        commands.append(f"UPDATE pham SET Color = '{new_color}' WHERE "
                        f"PhamID = '{pham_id}'")

    mysqldb.execute_transaction(engine, commands)


def fix_colored_orphams(engine):
    """
    Find any single-member phams which are colored as though they are
    multi-member phams (not #FFFFFF in pham.Color).
    :param engine: sqlalchemy Engine allowing access to the database
    :return:
    """
    query = "SELECT * FROM (SELECT g.PhamID, COUNT(GeneID) AS count, " \
            "p.Color FROM gene AS g INNER JOIN pham AS p ON g.PhamID " \
            "=p.PhamID GROUP BY PhamID) AS c WHERE Color != '#FFFFFF' " \
            "AND count = 1"

    results = mysqldb_basic.query_dict_list(engine, query)
    print(f"Found {len(results)} non-white orphams...")

    commands = []
    for dictionary in results:
        pham_id = dictionary["PhamID"]
        count = dictionary["count"]
        color = dictionary["Color"]
        new_color = "#FFFFFF"
        commands.append(f"UPDATE pham SET Color = '{new_color}' WHERE "
                        f"PhamID = '{pham_id}'")

    mysqldb.execute_transaction(engine, commands)


def preserve_phams(old_phams, new_phams, old_colors, new_genes):
    """
    Attempts to keep pham numbers consistent from one round of pham
    building to the next
    :param old_phams: the dictionary that maps old phams to their genes
    :param new_phams: the dictionary that maps new phams to their genes
    :param old_colors: the dictionary that maps old phams to colors
    :param new_genes: the set of previously un-phamerated genes
    :return:
    """
    final_phams = dict()
    final_colors = dict()

    outcount = 0
    total = len(old_phams)
    new_phams_copy = new_phams.copy()

    # Iterate through old and new phams
    for old_key in old_phams.keys():
        outcount += 1
        # print("Pham Name Conservation: {} / {}".format(outcount, total))

        old_pham = old_phams[old_key]

        if old_key in final_phams.keys():
            continue

        for new_key in new_phams_copy.keys():
            new_pham = new_phams_copy[new_key]

            if old_pham & new_pham == set():
                continue

            # Case 1 + 5 (Identity and Subtraction)
            if old_pham == new_pham:
                final_phams[old_key] = new_pham
                final_colors[old_key] = old_colors[old_key]
                new_phams_copy.pop(new_key)
                break

            # Case 2 and 4 (Addition and Join) - PHAM GREW
            elif new_pham - old_pham != set():

                # Case 2 and 4 (Addition and Join)
                if new_pham & new_genes != set():

                    # Case 4 - Join with new gene
                    if (new_pham - (new_pham & new_genes)) - old_pham != set():
                        break

                    # Case 2 - Addition with new gene
                    final_phams[old_key] = new_pham
                    final_colors[old_key] = old_colors[old_key]
                    new_phams_copy.pop(new_key)
                    break

                # Case 4 - Join without new gene
                else:
                    break

            # Case 3 - split - PHAM SHRANK, BUT NOT BY REMOVAL
            elif old_pham - new_pham != set():
                break

    final_phams[0] = "placeholder"
    highest_pham = max(map(int, final_phams.keys())) + 1
    final_phams.pop(0)

    # Reassign data for split or joined phams
    for key in new_phams_copy.keys():
        new_key = highest_pham
        highest_pham += 1

        final_phams[new_key] = new_phams_copy[key]

        # Multi-member pham gets non-white color
        if len(new_phams_copy[key]) > 1:
            h = s = v = 0
            while h <= 0:
                h = random.random()
            while s < 0.5:
                s = random.random()
            while v < 0.8:
                v = random.random()
            rgb = colorsys.hsv_to_rgb(h, s, v)
            rgb = (rgb[0] * 255, rgb[1] * 255, rgb[2] * 255)
            hexrgb = "#{:02x}{:02x}{:02x}".format(int(rgb[0]), int(rgb[1]),
                                                  int(rgb[2]))
            final_colors[new_key] = hexrgb.upper()
        # Orpham gets white
        else:
            final_colors[new_key] = '#FFFFFF'

    return final_phams, final_colors



"""Basic functions to interact with MySQL and manage databases."""

import subprocess

from pdm_utils.functions import basic

def drop_db(engine, database):
    """Delete a database.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param database: Name of the database to drop.
    :type database: str
    :returns: Indicates if drop was successful (0) or failed (1).
    :rtype: int
    """
    statement = f"DROP DATABASE {database}"
    try:
        engine.execute(statement)
    except:
        return 1
    else:
        return 0

def create_db(engine, database):
    """Create a new, empty database.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param database: Name of the database to create.
    :type database: str
    :returns: Indicates if create was successful (0) or failed (1).
    :rtype: int
    """
    statement = f"CREATE DATABASE {database}"
    try:
        engine.execute(statement)
    except:
        return 1
    else:
        return 0

def drop_create_db(engine, database):
    """Creates a new, empty database.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param database: Name of the database to drop and create.
    :type database: str
    :returns: Indicates if drop/create was successful (0) or failed (1).
    :rtype: int
    """
    # First, test if the database already exists within mysql.
    # If there is, delete it so that a new database is installed.
    databases = get_mysql_dbs(engine)
    if database in databases:
        result = drop_db(engine, database)
    else:
        result = 0
    if result == 0:
        result = create_db(engine, database)
    return result

def install_db(engine, schema_filepath):
    """Install a MySQL file into the indicated database.

    :param engine:
        SQLAlchemy Engine object able to connect to a MySQL databas.
    :type engine: Engine
    :param schema_filepath: Path to the MySQL database file.
    :type schema_filepath: Path
    :returns: Indicates if copy was successful (0) or failed (1).
    :rtype: int
    """
    cmd = mysql_login_command(engine.url.username,
                              engine.url.password,
                              engine.url.database)
    with schema_filepath.open("r") as fh:
        print("Installing database...")
        try:
            subprocess.check_call(cmd, stdin=fh)
        except:
            print(f"Unable to install {schema_filepath.name} in MySQL.")
            result = 1
        else:
            print("Installation complete.")
            result = 0
    return result

def copy_db(engine, new_database):
    """Copies a database.

    :param engine:
        SQLAlchemy Engine object able to connect to a MySQL database, which
        contains the name of the database that will be copied into
        the new database.
    :type engine: Engine
    :param new_database: Name of the new copied database.
    :type new_database: str
    :returns: Indicates if copy was successful (0) or failed (1).
    :rtype: int
    """
    if engine.url.database == new_database:
        print("Databases are the same so no copy needed.")
        result = 0
    else:
        dbs = get_mysql_dbs(engine)
        if new_database not in dbs:
            print(f"Unable to copy {engine.url.database} to "
                  f"{new_database} since {new_database} does not exist.")
            result = 1
        else:
            #mysqldump -u root -pPWD database1 | mysql -u root -pPWD database2
            cmd1 = mysqldump_command(engine.url.username,
                                     engine.url.password,
                                     engine.url.database)
            cmd2 = mysql_login_command(engine.url.username,
                                       engine.url.password,
                                       new_database)
            print("Copying database...")
            try:
                pipe_commands(cmd1, cmd2)
            except:
                print(f"Unable to copy {engine.url.database} to "
                      f"{new_database} in MySQL due to copying error.")
                result = 1
            else:
                print("Copy complete.")
                result = 0

    return result

def db_exists(engine, database):
    """Check if given name for a local MySQL database exists.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param database: The name of the database to check exists.
    :type database: str
    :returns: Returns whether the database exists
    :rtype: bool
    """
    exists_query = ("SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA "
                   f"WHERE SCHEMA_NAME = '{database}'")

    existing_db = scalar(engine, exists_query)
    return (not existing_db is None)

# TODO test.
def pipe_commands(command1, command2):
    """Pipe one command into the other."""
    # Per subprocess documentation:
    # 1. For pipes, use Popen instead of check_call.
    # 2. Call p1.stdout.close() to allow p1 to receive a SIGPIPE
    #    if p2 exits, which is called when used as a context manager.
    # communicate() waits for the process to complete.
    with subprocess.Popen(command1, stdout=subprocess.PIPE) as p1:
        with subprocess.Popen(command2, stdin=p1.stdout) as p2:
            p2.communicate()

def mysqldump_command(username, password, database):
    """Construct list of strings representing a mysqldump command."""
    # mysqldump -u root -pPWD database1 > database.sql
    # output filename is not needed since redirecting stdout.
    cmd = f"mysqldump -u {username} -p{password} {database}"
    cmd_list = cmd.split(" ")
    return cmd_list

def mysql_login_command(username, password, database):
    """Construct list of strings representing a mysql command."""
    # mysql -u root -pPWD database
    cmd = (f"mysql -u {username} -p{password} {database}")
    cmd_list = cmd.split(" ")
    return cmd_list


def get_mysql_dbs(engine):
    """Retrieve database names from MySQL.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :returns: Set of database names.
    :rtype: set
    """
    query = "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA"
    databases = query_set(engine, query)
    return databases


def get_tables(engine, database):
    """Retrieve tables names from the database.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :returns: Set of table names.
    :rtype: set
    """
    query = ("SELECT table_name FROM information_schema.tables "
             f"WHERE table_schema = '{database}'")
    tables = query_set(engine, query)
    return tables


def get_columns(engine, database, table_name):
    """Retrieve columns names from a table.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param database: Name of the database to query.
    :type database: str
    :param table_name: Name of the table to query.
    :type table_name: str
    :returns: Set of column names.
    :rtype: set
    """
    query = ("SELECT column_name FROM information_schema.columns WHERE "
              f"table_schema = '{database}' AND "
              f"table_name = '{table_name}'")
    columns = query_set(engine, query)
    return columns


# TODO test.
def query_set(engine, query):
    """Retrieve set of data from MySQL query.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param query: MySQL query statement.
    :type query: str
    :returns: Set of queried data.
    :rtype: set
    """
    result_list = engine.execute(query).fetchall()
    set_of_data = basic.get_values_from_tuple_list(result_list)
    return set_of_data

# TODO unittest.
def query_dict_list(engine, query):
    """Get the results of a MySQL query as a list of dictionaries.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param query: MySQL query statement.
    :type query: str
    :returns:
        List of dictionaries, where each dictionary represents a row of data.
    :rtype: list
    """
    result_list = engine.execute(query).fetchall()
    result_dict_list = []
    for row in result_list:
        row_as_dict = dict(row)
        result_dict_list.append(row_as_dict)
    return result_dict_list

def retrieve_data(engine, column=None, query=None, id_list=None):
    """Retrieve genome data from a MySQL database for a single genome.

    The query is modified to include one or more values.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param query:
        A MySQL query that selects valid, specific columns
        from the a valid table without conditioning on a specific column
        (e.g. 'SELECT Column1, Column2 FROM table1').
    :type query: str
    :param column:
        A valid column in the table upon which the query can be conditioned.
    :type column: str
    :param id_list:
        A list of valid values upon which the query can be conditioned.
        In conjunction with the 'column' parameter, the 'query' is
        modified (e.g. "WHERE Column1 IN ('Value1', 'Value2')").
    :type id_list: list
    :returns:
        A list of items, where each item is a dictionary of
        SQL data for each PhageID.
    :rtype: list
    """
    if (id_list is not None and len(id_list) > 0):
        query = query \
                + f" WHERE {column} IN ('" \
                + "','".join(id_list) \
                + "')"
    query = query + ";"
    result_dict_list = query_dict_list(engine, query)
    return result_dict_list


def get_distinct(engine, table, column, null=None):
    """Get set of distinct values currently in a MySQL database.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param table: A valid table in the database.
    :type table: str
    :param column: A valid column in the table.
    :type column: str
    :param null: Replacement value for NULL data.
    :type null: misc
    :returns: A set of distinct values from the database.
    :rtype: set
    """
    query = f"SELECT DISTINCT({column}) FROM {table}"
    result_set = query_set(engine, query)

    if None in result_set:
        result_set.remove(None)
        result_set.add(null)

    return result_set


def get_table_count(engine, table):
    """Get the current number of genomes in the database.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :returns: Number of rows from the phage table.
    :rtype: int
    """
    query = f"SELECT COUNT(*) FROM {table}"
    result_list = engine.execute(query).fetchall()
    count = result_list[0][0]
    return count

def get_first_row_data(engine, table):
    """Retrieves data from the first row of a table.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :returns: Dictionary where key = column name.
    :rtype: dict
    """
    query = f"SELECT * FROM {table}"
    result_dict_list = query_dict_list(engine, query)
    if len(result_dict_list) > 0:
        dict = result_dict_list[0]
    else:
        dict = {}
    return dict


def first(engine, executable, return_dict=True):
    """Execute a query and get the first row of data.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param executable: Input an executable MySQL query.
    :type executable: Select
    :type executable: str
    :param return_dict: Toggle whether execute returns dict or tuple.
    :type return_dict: Boolean
    :returns: Results from execution of given MySQL query.
    :rtype: dict
    :rtype: tuple
    """
    proxy = engine.execute(executable)
    first_row = proxy.first()
    if return_dict:
        first_row = dict(first_row)
    else:
        first_row = tuple(first_row)
    return first_row


def scalar(engine, executable):
    """Execute a query and get the first field.

    :param engine: SQLAlchemy Engine object able to connect to a MySQL database.
    :type engine: Engine
    :param executable: Input an executable MySQL query.
    :type executable: Select
    :type executable: str
    :returns: Scalar result from execution of given MySQL query.
    :rtype: int
    """
    proxy = engine.execute(executable)
    scalar = proxy.scalar()
    return scalar


def convert_for_sql(value, check_set=set(), single=True):
    """Convert a value for inserting into MySQL.

    :param value: Value that should be checked for conversion.
    :type value: misc
    :param check_set: Set of values to check against.
    :type check_set: set
    :param single: Indicates whether single quotes should be used.
    :type single: bool
    :returns:
        Returns either "NULL" or the value encapsulated in quotes
        ("'value'" or '"value"')
    :rtype: str
    """
    if value in check_set:
        value = "NULL"
    else:
        if single == True:
            value = f"'{value}'"
        else:
            value = f'"{value}"'
    return value



from pathlib import Path
import sys

import sqlalchemy
from getpass import getpass
from sqlalchemy import create_engine
from sqlalchemy import MetaData
from sqlalchemy.orm import sessionmaker
from sqlalchemy.engine.base import Engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.ext.automap import automap_base

from pdm_utils.functions import querying
from pdm_utils.functions import mysqldb_basic
from pdm_utils.functions import parsing

# -----------------------------------------------------------------------------
# GLOBAL VARIABLES
SUPPORTED_DIALECTS = ["mysql", "sqlite"]
SUPPORTED_DRIVERS = {"mysql": ["pymysql"],
                     "sqlite": ["pysqlite"]}

DEFAULT_DRIVER = {"mysql": "pymysql",
                  "sqlite": "pysqlite"}

CREDENTIALS_MSG = ("Credentials invalid and maximum login attempts reached. "
                   "Please check your MySQL credentials and try again.")

DATABASE_MSG = ("Unable to connect to database with valid credentials.\n"
                "Please check your SQL database access, "
                "and/or your database availability.")


class AlchemyHandler:
    def __init__(self, database=None, username=None, password=None,
                 dialect="mysql", driver=None):
        self._database = database
        self._username = username
        self._password = password
        # Username, password, and database are included for two reasons:
        # The ability to set credentials 'pythonically'
        # An understanding of whether the credentials are valid for the engine.

        self._engine = None
        self._metadata = None
        self._graph = None
        self._mapper = None
        self._session = None
        self.echo = False

        if dialect not in SUPPORTED_DIALECTS:
            raise NotImplementedError(
                            f"SQL SqlAlchemy dialect {dialect} not supported.")
        self._dialect = dialect

        if driver is None:
            driver = DEFAULT_DRIVER[dialect]

        if driver not in SUPPORTED_DRIVERS[dialect]:
            raise NotImplementedError(
                            f"SQL SqlAlchemy driver {driver} not supported.")
        self._driver = driver

        self.connected = False
        self.has_credentials = False

        if (username is not None) and (password is not None):
            self.has_credentials = True

        self.has_database = False
        self.connected_database = False
        self._databases = []

        if database is not None:
            self.has_database = True

# -----------------------------------------------------------------------------
# ALCHEMYHANDLER PROPERTIES
# -----------------------------------------------------------------------------

    @property
    def database(self):
        """Returns the AlchemyHandler's set database.

        :returns: Returns a copy of the database attribute.
        :rtype: str
        """
        database = self._database
        return database

    @database.setter
    def database(self, database):
        if database is None:
            self.has_database = False
            self.connected_database = False
            return

        if not isinstance(database, str):
            raise TypeError("Entered database name is not a string.")

        self._database = database

        self.has_database = True
        self.connected_database = False

    @property
    def username(self):
        """Returns the AlchemynHandler's set username.

        :returns: Returns a copy of the username attribute.
        :rtype: str
        """
        username = self._username
        return username

    @username.setter
    def username(self, username):
        if username is None:
            self.has_credentials = False
            self.connected = False
            return

        if not isinstance(username, str):
            raise TypeError("Entered username is not a string")

        self._username = username

        if self._password is not None:
            self.has_credentials = True
        self.connected = False

        self.clear()

    @property
    def password(self):
        """Returns the AlchemyHandler's set password.

        :returns: Returns a copy of the password attribute.
        :rtype: str
        """
        password = self._password
        return password

    @password.setter
    def password(self, password):
        if password is None:
            self.has_credentials = False
            self.connected = False
            return

        if not isinstance(password, str):
            raise TypeError("Entered password is not a string.")

        self._password = password

        if self._username is not None:
            self.has_credentials = True
        self.connected = False

        self.clear()

    @property
    def login_attempts(self):
        """Returns the AlchemyHandler's number of login attempts for login.

        :returns: Returns the number of login attempts for login.
        :rtype: str
        """
        login_attempts = self._login_attempts
        return login_attempts

    @property
    def engine(self):
        """Returns the AlchemyHandler's stored engine object.

        :returns: Returns the AlchemyHandler's stored engine object.
        :rtype: Engine
        """
        if self._engine is None:
            self.build_engine()

        engine = self._engine
        return engine

    @engine.setter
    def engine(self, engine):
        self.clear()

        if engine is None:
            self.connected = False
            self._engine = None
            return

        if not isinstance(engine, Engine):
            raise TypeError

        self._engine = engine
        self.extract_engine_credentials(engine)
        self.get_mysql_dbs()

    @property
    def URI(self):
        """Returns a SQLAlchemy URI string from stored credentials.
        """
        if not self.connected:
            return ""

        return self.construct_engine_string(username=self._username,
                                            password=self._password,
                                            database=self._database,
                                            dialect=self._dialect,
                                            driver=self._driver)

    @URI.setter
    def URI(self, URI):
        self.clear()

        if URI is None:
            self._engine = None
            self._username = None
            self._password = None
            self._database = None
            self.has_credentials = False
            self.has_database = False
            return

        if not isinstance(URI, str):
            raise TypeError

        self._engine = create_engine(URI)
        self.extract_engine_credentials(self._engine)

    @property
    def session(self):
        """Returns the AlchemyHandler's stored session object.
        """
        if self._session is None:
            self.build_session()

        session = self._session
        return session

    @property
    def metadata(self):
        """Returns the AlchemyHandler's stored metadata object.

        :returns: Returns the AlchemyHandler's stored engine object.
        :rtype: MetaData
        """
        if self._metadata is None:
            self.build_metadata()

        metadata = self._metadata
        return metadata

    @property
    def graph(self):
        """Returns the AlchemyHandler's stored graph object.

        :returns: Returns the AlchemyHandler's stored metadata graph object.
        :rtype: Graph
        """
        if self._graph is None:
            self.build_graph()

        graph = self._graph
        return graph

    @property
    def mapper(self):
        """Returns the AlchemyHandler's stored automapper object.

        :returns: Returns the AlchemyHandler's stored mapper object.
        """
        if self._mapper is None:
            self.build_mapper()

        mapper = self._mapper
        return mapper

    @property
    def databases(self):
        """Returns a copy of the databases available to the current credentials

        :returns: Returns the AlchemyHandler's available databases.
        :rtype: list[str]
        """
        databases = self._databases.copy()
        return databases

# -----------------------------------------------------------------------------
# CONNECTION METHODS
# -----------------------------------------------------------------------------

    def ask_database(self):
        """Ask for database input to store in AlchemyHandler.
        """
        if self._dialect == "mysql":
            self._database = input("MySQL database: ")

        elif self._dialect == "sqlite":
            self._database = str(Path(input("SQLite database: ")))

        self.has_database = True
        self.connected = False

    def ask_credentials(self):
        """Ask for username and password input to store in AlchemyHandler.
        """
        if self._dialect == "mysql":
            self._username = getpass(prompt="MySQL username: ")
            self._password = getpass(prompt="MySQL password: ")

        self.has_credentials = True
        self.connected = False

    def extract_engine_credentials(self, engine):
        """Extract username, password, and/or database from a SQLAlchemy engine.
        """
        url = engine.url
        self.connected = True

        self._username = url.username
        self._password = url.password
        self.has_credentials = True

        if url.database != "":
            self._database = url.database
            self.has_database = True
            self.connected_database = True

    def validate_database(self):
        """Validate access to database using stored SQL credentials.
        """
        if not self.connected:
            raise ValueError(
                    "AlchemyHandler currently not connected to MySQL.")
        if not self.has_database:
            raise AttributeError("No database in AlchemyHandler to validate")

        if self._dialect == "mysql":
            self.get_mysql_dbs()

            if self._database not in self._databases:
                raise MySQLDatabaseError("User does not have access to "
                                         f"database {self._database}")

        elif self._dialect == "sqlite":
            if not Path(self._database).is_file():
                raise SQLiteDatabaseError(
                                "Specified sqlite database does not exist.")

    def build_engine(self):
        """Create and store SQLAlchemy Engine object.
        """
        if self._dialect == "mysql":
            if not self.connected:
                if not self.has_credentials:
                    raise AttributeError(
                                "AlchemyHandler missing credentials.\n"
                                "Cannot connect to MySQL.")
                login_string = self.construct_engine_string(
                                            username=self._username,
                                            password=self._password,
                                            dialect=self._dialect,
                                            driver=self._driver)

                self.clear()

                self._engine = sqlalchemy.create_engine(login_string,
                                                        echo=self.echo)
                self._engine.connect()

                self.connected = True

                self.get_mysql_dbs()

        elif self._dialect == "sqlite":
            self.connected = True

        if self.has_database and (not self.connected_database):
            self.validate_database()

            login_string = self.construct_engine_string(
                                        username=self._username,
                                        password=self._password,
                                        database=self._database,
                                        dialect=self._dialect,
                                        driver=self._driver)

            self.clear()

            self._engine = sqlalchemy.create_engine(login_string,
                                                    echo=self.echo)
            self._engine.connect()

            self.connected = True
            self.connected_database = True

    def connect(self, ask_database=False, login_attempts=5, pipeline=False):
        """Ask for input to connect to MySQL and MySQL databases.

        :param ask_database: Toggle whether to connect to a database.
        :type ask_database: Boolean
        :param login_attempts: Set number of total login attempts.
        :type login_attempts: int
        """
        attempts = 0
        if self._dialect == "mysql":
            if self.has_credentials:
                try:
                    self.build_engine()
                except OperationalError:
                    pass
                except AttributeError:
                    pass
                except SQLCredentialsError:
                    pass
                except MySQLDatabaseError:
                    pass

            while(not self.connected and attempts < login_attempts):
                attempts += 1
                self.ask_credentials()

                try:
                    self.build_engine()
                except OperationalError:
                    pass
                except AttributeError:
                    pass
                except SQLCredentialsError:
                    pass
                except MySQLDatabaseError:
                    pass

            if not self.connected:
                if not pipeline:
                    raise SQLCredentialsError(CREDENTIALS_MSG)
                else:
                    print(CREDENTIALS_MSG)
                    sys.exit(1)

            if ask_database:
                try:
                    self.build_engine()
                except OperationalError:
                    pass
                except AttributeError:
                    pass
                except MySQLDatabaseError:
                    pass

                while(not self.connected_database and
                      attempts < login_attempts):
                    attempts += 1
                    self.ask_database()

                    try:
                        self.build_engine()
                    except OperationalError:
                        pass
                    except AttributeError:
                        pass
                    except MySQLDatabaseError:
                        pass

                if not self.connected_database:
                    if not pipeline:
                        raise MySQLDatabaseError(DATABASE_MSG)
                    else:
                        print(DATABASE_MSG)
                        sys.exit(1)

        elif self._dialect == "sqlite":
            self.connected = True
            self.has_credentials = True

            if self.has_database:
                try:
                    self.build_engine()
                except OperationalError:
                    pass
                except AttributeError:
                    pass
                except SQLiteDatabaseError:
                    pass

            while(not self.connected_database and
                  attempts < login_attempts):
                attempts += 1
                self.ask_database()

                try:
                    self.build_engine()
                except OperationalError:
                    pass
                except AttributeError:
                    pass
                except SQLiteDatabaseError:
                    pass

            if not self.connected_database:
                if not pipeline:
                    raise SQLiteDatabaseError(DATABASE_MSG)
                else:
                    print(DATABASE_MSG)
                    sys.exit(1)

    def construct_engine_string(self, dialect="mysql", driver="pymysql",
                                username="", password="", database=""):
        """Construct a SQLAlchemy engine URL.

        :param dialect: Type of SQL database.
        :type dialect: str
        :param driver: Name of the Python DBAPI used to connect.
        :type driver: str
        :param username: Username to login to SQL database.
        :type username: str
        :param password: Password to login to SQL database.
        :type password: str
        :param database: Name of the database to connect to.
        :type database: str
        :returns: URI string to create SQLAlchemy engine.
        :rtype: str
        """
        if driver != "":
            dbapi = "+".join([dialect, driver])
        else:
            dbapi = dialect

        if dialect in ("mysql", "postgresql"):
            engine_string = (f"{dbapi}://{username}:"
                             f"{password}@localhost/{database}")
        elif dialect == "sqlite":
            engine_string = (f"{dbapi}:///{database}")
        else:
            raise NotImplementedError(
                            f"SQL SqlAlchemy dialect {dialect} not supported.")
        return engine_string

    def get_mysql_dbs(self):
        """Retrieve database names from MySQL.

        :returns: List of database names.
        :rtype: list
        """
        if not self.connected:
            raise ValueError(
                        "AlchemyHandler currently not connected to MySQL.")
        databases = mysqldb_basic.get_mysql_dbs(self._engine)
        self._databases = list(databases)

    def clear(self):
        """Clear properties tied to MySQL credentials/database.
        """
        if self._engine is not None:
            self._engine.dispose()
        self._engine = None

        self._metadata = None
        self._graph = None
        self._mapper = None

        if self._session is not None:
            self._session.close()
        self._session = None

        self.connected = False
        self.connected_database = False
        self._databases = []

# -----------------------------------------------------------------------------
# SQLALCHEMY-RELATED OBJECT GENERATION METHODS

    def build_metadata(self):
        """Create and store SQLAlchemy MetaData object.
        """
        if not self.has_database:
            self.ask_database()

        if not self.connected_database:
            self.build_engine()

        self._metadata = MetaData(bind=self._engine)
        self._metadata.reflect()

    def build_session(self):
        """Create and store SQLAlchemy Session object.
        """
        if not self.has_database:
            self.ask_database()

        if not self.connected:
            self.build_engine()

        if self._session is not None:
            self._session.close()

        session_maker_obj = sessionmaker(bind=self._engine)
        self._session = session_maker_obj()

    def build_graph(self):
        """Create and store SQLAlchemy MetaData related NetworkX Graph object.
        """
        if self._metadata is None:
            self.build_metadata()

        self._graph = querying.build_graph(self._metadata)

    def build_mapper(self):
        """Create and store SQLAlchemy automapper Base object.
        """
        if self._metadata is None:
            self.build_metadata()

        self._mapper = automap_base(metadata=self._metadata)
        self._mapper.prepare()

    def build_all(self):
        """Create and store all relevant SQLAlchemy objects.
        """
        self.build_session()
        self.build_graph()
        self.build_mapper()

# -----------------------------------------------------------------------------
# SQLALCHEMY QUALITY-OF-LIFE FUNCTIONS

    def get_map(self, table):
        """Get SQLAlchemy ORM map object.
        """
        if self._mapper is None:
            self.build_mapper()

        table = parsing.translate_table(self._metadata, table)
        return self._mapper.classes[table]


class SQLCredentialsError(Exception):
    pass


class MySQLDatabaseError(Exception):
    pass


class SQLiteDatabaseError(Exception):
    pass
